#include <linux/ioctl.h>
#include <linux/cdev.h>

#define QUASAR_MAX_SPANS 16
#define QUASAR_MAX_CHANS 32

#define PMC_OLD_FIRMWARE	0x1113		// for Quasar8
#define SUBSYSTEM_EC	0x0001
#define SUBSYSTEM_WD	0x0002

#define QUASAR2012(pdev) ((pdev->device == PCI_DEVICE_ID_QUASAR2012) || (pdev->device == PCI_DEVICE_ID_QUASAR8LP))
#define QUASAR8(pdev) (pdev->device == PCI_DEVICE_ID_QUASAR8)
#define QUASARM(pdev) (pdev->device == PCI_DEVICE_ID_QUASARM)
#define QUASARMICRO(pdev) (pdev->device == PCI_DEVICE_ID_QUASARMICRO)
#define QUASAR16(pdev) ((pdev->device == PCI_DEVICE_ID_QUASAR16)||(pdev->device == PCI_DEVICE_ID_QUASAR16LP))

enum bar0_regs_t {
    BAR0_IFR = 0x00,
    BAR0_IMR = 0x01,
    BAR0_CTRL = 0x02,
    BAR0_TXAO = 0x04,
    BAR0_LINE = 0x05,
    BAR0_BSLS = 0x06,
    BAR0_DMA_CT_CFG = 0x08,
    BAR0_DMACR = 0x09,
    BAR0_DMA_ADDR = 0x0a,
    BAR0_DMA_WADDR = 0x0b,
    BAR0_DMA_RADDR = 0x0c,
    BAR0_HW_FW_REV = 0x0d,
    BAR0_FIRM_CTRL = 0x0e,
    BAR0_FIRM_PASS = 0x0f,
    BAR0_RTC = 0x10,
    BAR0_LIU_CTRL = 0x12,
    BAR0_I2C = 0x13,
    BAR0_PCI_LOAD = 0x20,
    BAR0_PMC_LOAD = 0x21,
    BAR0_DMA_READ_BUFF = 0x100,
    BAR0_DMA_WRITE_BUFF = 0x200,
    BAR0_CT_E1_SWITCH = 0x1000,
    BAR0_E1_CT_SWITCH = 0x2000,
};

struct pmc_e1_cas_data
{
    unsigned char chan_num;
    unsigned char cdata[16];
};

struct quasar
{
	struct pci_dev			*pci_device;	//coresponding pci_dev struct
	struct list_head		dev_node;	//list of devices, may be we don't need it now...
	struct cdev				dev_file;	//descriptop file
	spinlock_t              dev_lock;
	unsigned char           dev_number;
	volatile unsigned int	*bar0;
	volatile unsigned int	*bar1;		//pointers to map device's BARS
	unsigned char           e1_amount;      //number of last e1 if
	unsigned char			initialized:1;
	unsigned char			opened:1;
	unsigned char			registered:1;
	char                    name[30];
	unsigned char           pos_in_write_buf;
	unsigned char           have_liu[QUASAR_MAX_SPANS];
	void *                  priv_data;      //hook for dahdi
	unsigned int            test_sig;
	struct device     		*pmc_class_dev;
	struct dahdi_device 	*ddev;
	int 					ec,wd;
	dma_addr_t				readdma;
  	dma_addr_t				writedma;
	void 					*hal_ops;
	int 					channels;
};
struct  quasar_span
{
    struct dahdi_span span;
    struct dahdi_chan _chans[QUASAR_MAX_CHANS];
    struct dahdi_chan *chans[QUASAR_MAX_CHANS];
    int spannumber;		                        /* span number in card */
    int syncprior;			                /* use or not this span for syncronization, and syncsource priority */
    void * pvt;
    int clear_counters;
    unsigned long frok;					/* Frame is stable frok ticks*/
    unsigned char psr;                                  /* LXT360 PSR register */
    unsigned char ec_chunk1[QUASAR_MAX_CHANS][DAHDI_CHUNKSIZE];          /* first EC chunk buffer */
    unsigned char ec_chunk2[QUASAR_MAX_CHANS][DAHDI_CHUNKSIZE];          /* second EC chunk buffer */
	int	unframed;
};

struct quasar_dahdi
{
  volatile unsigned int *writechunk;			/* Double-word aligned write memory */
  volatile unsigned int *readchunk;                     /* Double-word aligned read memory */
  dma_addr_t 	readdma;
  dma_addr_t	writedma;
  struct quasar_span    *spans[QUASAR_MAX_SPANS];
  unsigned char num_spans;		                        /* how many spans this card has */
  unsigned char qn_span;                                /*For fast check during the interrupt*/
  char registered;
  void * pvt;
  u32 ovfl_counter;                                     /* keep old value of overflow counter, used in e12zap_delivery */
  char check;
  unsigned int keep;					/* used for checking interrupt misses */
  int intcount;
  int irqmisses;        				//digium defined it as signed...
#ifdef QUASAR_TASKLET
  int taskletrun;
  int taskletsched;
  char taskletpending;
  int taskletexec;
  int txerrors;
  struct tasklet_struct quasar_tlet;
#endif
};

struct quasar_hal_ops {
	void (*init_core_regs)(struct quasar *pq);
	void (*init_switch_map)(struct quasar *pq);
	void (*set_hw_options)(struct quasar *pq, int noec);
	void (*init_framer_regs)(struct quasar *pq);
	int (*ioctl)(struct inode *inode_p, struct file *file_p,
			unsigned int cmd, unsigned long arg);
	void (*quasar_find_master)(struct quasar_dahdi *pd);
	void (*check_alarm)(struct quasar_dahdi *pd);
	void (*clearstat)(struct quasar_dahdi * pd);
	int (*maint)(struct dahdi_span *span, int cmd);
	int (*rbsbits)(struct dahdi_chan *chan, int bits);
	int (*shutdown)(struct dahdi_span *span);
	int (*startup)(struct dahdi_span *span);
	void (*stop_dma)(struct quasar *pq);
	void (*rbs_delivery)(struct quasar *pq);
	void (*copy_chunk)(struct quasar * pq);
	int (*dacs) (struct dahdi_chan *dst, struct dahdi_chan *src);
};

#define QUASAR_HAL(x) ((struct quasar_hal_ops *)(pq->hal_ops))->x

#define getp_pmc_struct(pdev) list_entry(pdev, quasar, pci_dev)

/* Quasar8, Quasar2012 switch table */
#define CTB_TO_E1_TABLE(ct_ch, ct_ts) ( ct_ts | ( ct_ch << 8) )
#define E1_TO_CTB_TABLE(e1_ch, e1_ts) ( ( e1_ts | 0x00008000) | (e1_ch  << 8) )


#define PMC_BIT_(a)  (1<<a)

/* CTRL register - quasar8 */
#define CTRL_ALTRES PMC_BIT_(31)
#define CTRL_SDIOD1 PMC_BIT_(30)
#define CTRL_SDIOD2 PMC_BIT_(29)
#define CTRL_CT_TO_EN PMC_BIT_(27)
#define CTRL_CT_TI_EN PMC_BIT_(26)
#define CTRL_DMARST PMC_BIT_(25)
#define CTRL_PTENOVR PMC_BIT_(24)
#define CTRL_MASL PMC_BIT_(23)
#define CTRL_ABSEL PMC_BIT_(22)
#define CTRL_MTRESET PMC_BIT_(21)
#define CTRL_NR1USE PMC_BIT_(20)
#define CTRL_NR2USE PMC_BIT_(19)
#define CTRL_NR1EN PMC_BIT_(18)
#define CTRL_NR2EN PMC_BIT_(17)
#define CTRL_NRFREQ PMC_BIT_(16)
#define CTRL_CONS_DEL PMC_BIT_(15)
#define CTRL_USESYNC PMC_BIT_(11)
#define CTRL_SRST PMC_BIT_(10)
#define CTRL_FSEL PMC_BIT_(9)
#define CTRL_FLOCK PMC_BIT_(8)
#define CTRL_FREE PMC_BIT_(7)
#define CTRL_HLD PMC_BIT_(6)
#define CTRL_LOCK PMC_BIT_(5)

/* CTRL register - quasarm */
#define CTRL_TMS_DF PMC_BIT_(21)
#define CTRL_CLR_ZERO PMC_BIT_(20)
#define CTRL_LOF_DIOD PMC_BIT_(14)
#define CTRL_WDOG PMC_BIT_(8)

/* CTRL register - quasar2012 */
#define CTRL_DSP_RST PMC_BIT_(16)


//definitions for quasar-m framer status
#define FRS_FRAME_SYNC   (1 << 23)
#define FRS_MFRAME_SYNC  (1 << 22)
#define FRS_CRC_SYNC     (1 << 21)
#define FRS_FRAME_ERR    (1 << 20)
#define FRS_CRC4_ERR     (1 << 19)
#define FRS_RED          (1 << 18)
#define FRS_AIS          (1 << 17)
#define FRS_MFRAI        (1 << 16)
#define FRS_RAI_RX       (1 << 15)
#define FRS_EBIT_RX1     (1 << 14)
#define FRS_EBIT_RX0     (1 << 13)
#define FRS_SA_BITS4     (1 << 12)
#define FRS_SA_BITS3     (1 << 11)
#define FRS_SA_BITS2     (1 << 10)
#define FRS_SA_BITS1     (1 << 9)
#define FRS_SA_BITS0     (1 << 8)
#define FRS_SLIP_ERR     (1 << 4)
#define FRS_SLIP_DIR     (1 << 3)
#define FRS_LOS          (1 << 2)
#define FRS_HDB3_ERR     (1 << 1)
#define FRS_AMI_ERR       1

#define FRC1_ADDR_OFFSET 0
#define FRS_ADDR_OFFSET 2
#define FRS_HDB3ERR_OFFSET 4
#define FRS_FRAMEERR_OFFSET 5
#define FRS_CRC4ERR_OFFSET 6
#define FRS_EBITERR_OFFSET 7
#define SWITCH_TABLE_OFFSET 0x2000
#define INPUT_TS_SIGNATURE(LINE, TS) ( (LINE << 8) | TS)

//definitions for CAS related registers
#define CAS_BIT_BUFFER_OFFSET 0x10
#define CAS_Y_BIT 	4
#define CAS_X1_BIT 	8
#define CAS_X2_BIT 	2
#define CAS_X3_BIT 	1
#define CAS_RX_Y_BIT 	(4 << 8)
#define CAS_RX_X1_BIT 	(8 << 8)
#define CAS_RX_X2_BIT 	(2 << 8)
#define CAS_RX_X3_BIT 	(1 << 8)



//definitions for framer configuration
#define FRC_ACRC      (1 << 26)
#define FRC_ARAI      (1 << 25)
#define FRC_AEBIT     (1 << 24)
#define FRC_RAI_TX    (1 << 23)
#define FRC_EBIT_TX1  (1 << 22)
#define FRC_EBIT_TX0  (1 << 21)
#define FRC_SA_BITS4  (1 << 20)
#define FRC_SA_BITS3  (1 << 19)
#define FRC_SA_BITS2  (1 << 18)
#define FRC_SA_BITS1  (1 << 17)
#define FRC_SA_BITS0  (1 << 16)
#define FRC_TXAO      (1 << 12)
#define FRC_FR_RX_EN  (1 << 11)
#define FRC_FR_TX_EN  (1 << 10)
#define FRC_CRC4_TX_EN    (1 << 9)
#define FRC_MFR_TX_EN     (1 << 8)
#define FRC_SHORT_HAUL_EN (1 << 7)
#define FRC_TXEN          (1 << 6)
#define FRC_HDB3_RX_EN    (1 << 5)
#define FRC_HDB3_TX_EN    (1 << 4)
#define FRC_MLBK          (1 << 2)
#define FRC_DBLK          (1 << 1)
#define FRC_RLBK           1


//definitions for LIU bitfields
#define LR_CR1_JASEL0     (1 << 6)
#define LR_CR1_UNIEB      (1 << 4)
#define LR_CR1_ENCENB     (1 << 5)
#define LR_CR1_EC1         1
#define LR_CR1_EC4        (1 << 3)
#define LR_CR4_CODEV       1
#define LR_CR4_ZEROV       2

#define LR_CR2_ERLOOP	  (1 << 0)
#define LR_CR2_ELLOOP	  (1 << 1)
#define LR_CR2_EALOOP	  (1 << 2)
#define LR_CR2_ENLOOP	  (1 << 3)
#define LR_CR2_ETAOS      (1 << 4)
#define LR_CR2_EPAT0      (1 << 5)
#define LR_CR2_EPAT1      (1 << 6)
#define LR_CR2_RESET      (1 << 7)

#define LR_PSR_LOS	  (1 << 0)
#define LR_PSR_NLOOP	  (1 << 1)
#define LR_PSR_AIS	  (1 << 2)
#define LR_PSR_QRSS	  (1 << 3)
#define LR_PSR_RESERV4	  (1 << 4)
#define LR_PSR_DFMO	  (1 << 5)
#define LR_PSR_BIST	  (1 << 6)
#define LR_PSR_RESERV7	  (1 << 7)

//LIU quasar
#define LIU_N (1 << 10)
#define LIU_CS (1 << 9)
#define LIU_TR (1 << 8)
#define LIU_WRITE 0
#define LIU_READ 1

//LIU LXT

#define LR_CR1 0
#define LR_CR2 1
#define LR_CR3 2
#define LR_ICR 3
#define LR_TSR 4
#define LR_PSR 5
#define LR_ESR 6
#define LR_CR4 7

int acc_lxt_reg(struct quasar * pq ,unsigned char chan_num, unsigned char addr, unsigned char mode, unsigned char *data);


#define PMC_BIT_(a)	(1<<a)
#define PMC_GBIT(a)	((a) & 0x000000ff)
#define PMC_GPAGE(a) 	(((a)>>16) & 0x000000ff)
#define PMC_GREG(a)  	(((a)>>8) & 0x000000ff)
#define PMC_PAGE_REG_BIT(a,b,c) (a<<16|b<<8|1<<c)
#define PMC_PAGE_REG(a,b) 	(a<<16|b<<8)

/** Hardware positions of configurational bits. */
enum cfg0_hp_t {
	CFG0_CAS_HP = PMC_PAGE_REG_BIT(1,0x14,5),	/** CAS  (1=CSS/0=CAS)   1-14-5*/
	CFG0_XMIT_TR_HP = PMC_PAGE_REG_BIT(1,0x12,2),  	/** Transparent on transmit 1=transparent/0=framing on 1-12-2*/
	CFG0_RCV_TR_HP = PMC_PAGE_REG_BIT(1,0x12,3), 	/** Transparent on receive  1-12-3*/
	CFG0_MASTER_HP = PMC_BIT_(5),  			/** Master/Slave = 1/0 on E1*/
	CFG0_RCV_HDB3_HP = PMC_PAGE_REG_BIT(1,0x15,7),	/**Enable HDB encoding on RX 1-15-7*/
	CFG0_XMIT_HDB3_HP = PMC_PAGE_REG_BIT(1,0x15,5),	/**Enable HDB encoding on TX 1-15-5*/
};

/** International/National bits */
enum cfg1_hp_t {
	CFG1_NAT0_HP = PMC_PAGE_REG_BIT(1,0x17,4),   /**Sa4 1-17-4 */
	CFG1_NAT1_HP = PMC_PAGE_REG_BIT(1,0x17,3),   /**Sa5 1-17-3 */
	CFG1_NAT2_HP = PMC_PAGE_REG_BIT(1,0x17,2),   /**Sa6 1-17-2 */
	CFG1_NAT3_HP = PMC_PAGE_REG_BIT(1,0x17,1),   /**Sa7 1-17-1 */
	CFG1_NAT4_HP = PMC_PAGE_REG_BIT(1,0x17,0),   /**Sa8 1-17-0 */
	CFG1_INAT0_HP = PMC_PAGE_REG_BIT(1,0x12,0),  /** 1-12-0*/
	CFG1_INAT1_HP = PMC_PAGE_REG_BIT(1,0x12,1),  /** 1-12-1*/
};

enum cfg2_hp_t {
	CFG2_RAI_AUTO_ON_HP = PMC_PAGE_REG_BIT(1,0x10,4),  /** RAI: Auto or ON: 0=Auto/1=ON  1-10-4*/
	CFG2_RAI_ENA_HP = PMC_PAGE_REG_BIT(1,0x16,5),      /** RAI: 0=normal/1=transmit alarm, ignoring in Auto mode 1-16-5*/
	CFG2_RCAS_AUTO_ON_HP = PMC_PAGE_REG_BIT(1,0x10,3), /** Remote CAS: 0=Auto/0=ON 1-10-3*/
	CFG2_RCAS_ENA_HP = PMC_PAGE_REG_BIT(1,0x13,2),     /** Enable remote CAS (Y bit): 0=normal/1=LOM, ignoring in Auto mode 1-13-2*/
	CFG2_X1_HP = PMC_PAGE_REG_BIT(1,0x13,3),           /** X1 multiframe 1-13-3*/
	CFG2_X2_HP = PMC_PAGE_REG_BIT(1,0x13,1),           /** X2 multiframe 1-13-1*/
	CFG2_X3_HP = PMC_PAGE_REG_BIT(1,0x13,0),           /** X3 multiframe 1-13-0*/
};

/**CRC4 related */
enum cfg3_hp_t {
	CFG3_CRC4_TR_HP = PMC_PAGE_REG_BIT(1,0x10,6),	  /** CRC4 transparent regeneration on 1-10-6*/
	CFG3_CRC4_DIS_HP = PMC_PAGE_REG_BIT(1,0x10,2),    /** CRC4 1=disable/0=enable  1-10-2*/
	CFG3_CRC4_MANUAL_HP = PMC_PAGE_REG_BIT(1,0x10,5), /**CRC4 Automatic CRC internetworking 0=on/1=off 1-10-5*/
};

/** 4 testing */
enum cfg4_hp_t {
	CFG4_AIS_HP = PMC_PAGE_REG_BIT(1,0x11,4),      /** TxAO 1=disable/0=enable 1-11-4*/
	CFG4_LLOOP_HP = PMC_PAGE_REG_BIT(1,0x15,6),    /** Metallic loopback (shleif)  1=loop/0=normal 1-15-6*/
	CFG4_RLOOP_HP = PMC_PAGE_REG_BIT(1,0x15,2),    /** Remote loopback (echo) 1=loop/0=normal  1-15-2*/
	CFG4_TX_DIS_HP = PMC_PAGE_REG_BIT(2,0x10,5),   /** 1=Disable transmition on the board level, use to emulate LOS, transmition enabled by default*/
	CFG4_PLOOP_HP = PMC_PAGE_REG_BIT(1,0x15,0),    /** Payload loopback 1=loop/0=normal 1-15-0 */
};

enum err_emu0_hp_t {
	ERR_EMU0_LOF_NFAS_HP = PMC_PAGE_REG_BIT(1,0x19,4), /** Emulate single frame violation (1=generate error)  1-19-4*/
	ERR_EMU0_LOF_FAS_HP = PMC_PAGE_REG_BIT(1,0x19,5),  /** Emulate single frame violation (1=generate error) 1-19-5*/
	ERR_EMU0_CRC4_HP = PMC_PAGE_REG_BIT(1,0x19,6),     /** Emulate single CRC4 error (1=generate error)  1-19-6*/
};

/**Misc status bits*/
enum stat0_hp {
	STAT0_LOF_HP = PMC_PAGE_REG_BIT(3,0x10,7),	/** LOF signal 1-loss/0-ok 3-10-7*/
	STAT0_CAS_LOM_HP = PMC_PAGE_REG_BIT(3,0x10,6),	/** CAS LOM signal 1-loss/0-ok 3-10-6*/
	STAT0_CRC4_SYNC_HP = PMC_PAGE_REG_BIT(3,0x10,5),/** CRC4 syncronisation 1-loss/0-ok 3-10-5*/
	STAT0_E_BIT1_HP = PMC_PAGE_REG_BIT(3,0x10,3),	/** bit E state 1=error/0=ok !(3-10-3 &  3-10-4)*/
	STAT0_E_BIT2_HP = PMC_PAGE_REG_BIT(3,0x10,4),
	STAT0_CRC4_1_HP = PMC_PAGE_REG_BIT(3,0x11,6),	/** CRC4 state  1=error/0=ok  (3-11-6 | 3-11-7)*/
	STAT0_CRC4_2_HP = PMC_PAGE_REG_BIT(3,0x11,7),
	STAT0_LOS_HP = PMC_PAGE_REG_BIT(3,0x11,4) ,	/** LOS signal 1-loss/0-ok 3-11-4*/
	STAT0_AIS_HP = PMC_PAGE_REG_BIT(3,0x11,2),	/** AIS state 1=valid AIS or all ones/0=valid AIS is not being received 3-11-2*/
	STAT0_RAIS_HP = PMC_PAGE_REG_BIT(3,0x11,1),	/** RAIS state 1=error/0=normal 3-11-1*/

};

/** International/National bits */
enum stat1_hp {
	STAT1_NAT0_HP = PMC_PAGE_REG_BIT(3,0x18,4),      /**Sa4 3-18-4 */
	STAT1_NAT1_HP = PMC_PAGE_REG_BIT(3,0x18,3),      /**Sa5 3-18-3 */
	STAT1_NAT2_HP = PMC_PAGE_REG_BIT(3,0x18,2),      /**Sa6 3-18-2 */
	STAT1_NAT3_HP = PMC_PAGE_REG_BIT(3,0x18,1),      /**Sa7 3-18-1 */
	STAT1_NAT4_HP = PMC_PAGE_REG_BIT(3,0x18,0),      /**Sa8 3-18-0 */
	STAT1_INAT1_HP = PMC_PAGE_REG_BIT(3,0x18,7),     /**3-18-7*/
	STAT1_INAT0_HP = PMC_PAGE_REG_BIT(3,0x15,7),     /**3-15-7*/
	STAT1_SLIP_HP = PMC_PAGE_REG_BIT(4,0x12,1),      /**Slip occured  1=Slip, auto clearing after reading 4-12-1 */
};

/**Multifreim X Y bits*/
enum stat2_hp {
	STAT2_X1_HP = PMC_PAGE_REG_BIT(3,0x19,3),  /**3-19-3*/
	STAT2_Y_HP = PMC_PAGE_REG_BIT(3,0x19,2),   /**3-19-2*/
	STAT2_X2_HP = PMC_PAGE_REG_BIT(3,0x19,1),  /**3-19-1*/
	STAT2_X3_HP = PMC_PAGE_REG_BIT(3,0x19,0),  /**3-19-0*/
};

/** Overrun bits */
enum ovr_mgmt_hp {
	OVR_HDB3_HP = PMC_PAGE_REG_BIT(4,0x1f,3),  /**4-1f-3*/
	OVR_FAS_HP =  PMC_PAGE_REG_BIT(4,0x1f,7),  /**4-1f-7*/
	OVR_CRC4_HP = PMC_PAGE_REG_BIT(4,0x1f,6),  /**4-1f-6*/
};

/* Error counters */
enum err_counters {
	CNT_FAS_HP = PMC_PAGE_REG_BIT(4,0x13,0),
	CNT_BPVHI_HP = PMC_PAGE_REG_BIT(4,0x16,0),
	CNT_BPVLO_HP = PMC_PAGE_REG_BIT(4,0x17,0),
	CNT_CRC4HI_HP = PMC_PAGE_REG_BIT(4,0x18,0),
	CNT_CRC4LO_HP = PMC_PAGE_REG_BIT(4,0x19,0),
};

/** Selection of mode & a/b select*/
enum mode1_hp {
	MODE1_AB_HP = PMC_BIT_(22),           /**0=use A signals/1=use B signals*/
	MODE1_MA_SL_SEL_HP = PMC_BIT_(23),    /**0=slave on CT-BUS, 1=Master on CT-BUS*/
	MODE1_NF1_GEN_ENA_HP = PMC_BIT_(18) , /**1=Enable to generate Netref1*/
	MODE1_NF2_GEN_ENA_HP =  PMC_BIT_(17), /**1=Enable to generate Netref2*/
	MODE1_NF1_USE_ENA_HP =  PMC_BIT_(20), /**1=Enable to use  Netref1 for syncronization*/
	MODE1_NF2_USE_ENA_HP =  PMC_BIT_(19), /**1=Enable to use  Netref2 for syncronization*/
	MODE1_FLOCK =  PMC_BIT_(8), 	      /**  */
	MODE1_FREE =  PMC_BIT_(7), 	      /**1=Using Internal synchronization */
	MODE1_HOLD =  PMC_BIT_(6), 	      /**1=Holding last derived synchronization */
};

#define DMA_IFQ0    0x0
#define DMA_IFQ1    0x10
#define DMA_IFQ2    0x20
#define DMA_IFQ3    0x30
#define DMA_SIZE1   0x01
#define DMA_SIZE0   0x00
#define DMA_WENA    0x80
#define DMA_RENA    0x40

#define module_printk(level, fmt, args...) printk(level "%s: " fmt, THIS_MODULE->name, ## args)
#ifdef PMC_E1_DEBUG
# define module_dprintk(level, fmt, args...) \
    module_printk(level, fmt, ## args)
# define dprintk(level, fmt, args...) \
    printk(level fmt, ## args)
#else
# define module_dprintk(level, fmt, args...)
# define dprintk(level, fmt, args...)
#endif


/* === Inline functions === */


/* Quasar8  framer read/write (MT9076) */
/* TODO: BUG() if bad parameters. spanno 0..7, page 0..0x18, reg 0..0x1f */
static inline unsigned long _framer_read(struct quasar *dev,
					 int spanno, int page, int reg)
{
	dev->bar1[spanno * 0x40] = page&0xFF;
	return dev->bar1[spanno * 0x40 + (reg&0x1F)] & 0xFF;
}

/* Read framer register (spinlock'ed) */
static inline unsigned long framer_read(struct quasar *dev,
					int spanno, int page, int reg)
{
	unsigned long flags, r;

	spin_lock_irqsave(&dev->dev_lock, flags);
	r = _framer_read(dev, spanno, page, reg);
	spin_unlock_irqrestore(&dev->dev_lock, flags);

	return r;
}
#define framer_read_pr(a,b,c) framer_read(a,b,PMC_GPAGE(c),PMC_GREG(c))

static inline void _framer_write(struct quasar *dev,
				 int spanno, int page, int reg, unsigned long v)
{
	dev->bar1[spanno * 0x40] = page&0xFF;
	dev->bar1[spanno * 0x40 + (reg&0x1F)] = v&0xFF;
}

/* Write framer register (spinlock'ed) */
static inline void framer_write (struct quasar *dev,
				 int spanno, int page, int reg, unsigned long v)
{
	unsigned long flags;

	spin_lock_irqsave(&dev->dev_lock, flags);
	_framer_write(dev, spanno, page, reg, v);
	spin_unlock_irqrestore(&dev->dev_lock, flags);
}
#define framer_write_pr(a,b,c,d) framer_write(a,b,PMC_GPAGE(c),PMC_GREG(c),d)

/* Check framer register (spinlock'ed @ framer_read ) */
static inline int framer_checkbit(struct quasar *dev,
				  int spanno, unsigned long pmc_page_reg)
{
	return framer_read(dev, spanno, PMC_GPAGE(pmc_page_reg),
			   PMC_GREG(pmc_page_reg)) & PMC_GBIT(pmc_page_reg);
}

static inline void framer_setbit(struct quasar *dev,
				 int spanno, unsigned long pmc_page_reg) {
	unsigned long tmp;

	tmp = framer_read_pr(dev, spanno, pmc_page_reg);
	tmp |= PMC_GBIT(pmc_page_reg);
	framer_write_pr(dev, spanno, pmc_page_reg, tmp);

}

static inline void framer_clrbit(struct quasar *dev,
				 int spanno, unsigned long pmc_page_reg) {
	unsigned long tmp;

	tmp = framer_read_pr(dev, spanno, pmc_page_reg);
	tmp &= ~PMC_GBIT(pmc_page_reg);
	framer_write_pr(dev, spanno, pmc_page_reg, tmp);
}

/* Quasar m / Quasar 2012 framer read/write */

static inline unsigned long pmc_reg_read (struct quasar *pq,
			unsigned int spanno, unsigned int offset)
{
    return pq->bar1[spanno * 0x40 + offset];
}

static inline unsigned int pmc_reg_check (struct quasar *pq,
			unsigned int spanno, unsigned int offset, unsigned int mask)
{
    return pmc_reg_read(pq, spanno, offset)
	& mask;
}

static inline void pmc_reg_write (struct quasar *pq,
			unsigned int spanno, unsigned int offset, unsigned int v)
{
    pq->bar1[spanno * 0x40 + offset] = v;
}
