#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/ioctl.h>
#include <linux/poll.h>
#include <linux/cdev.h>
#include <linux/device.h>
#if defined(HAVE_UNLOCKED_IOCTL) && defined(CONFIG_BKL)                                                                                                         
# include <linux/smp_lock.h> /* lock_kernel */
#endif
#include <linux/version.h>

#include "dahdi/kernel.h"

#include "dver.h"
#include "kal.h"
#include "version.h"
#include "quasar_conf.h"
#include "quasar.h"

#define ECFG_PROTO_VER
#include "pmc_e1_if.h"

#ifdef HAVE_SPAN_COUNT
# define DAHDI_COUNT(a,b,c) a.c
#else
# define DAHDI_COUNT(a,b,c) a.b
#endif //HAVE_DAHDI_COUNT


static void init_core_regs_quasarm(struct quasar * pq)
{
	unsigned int volatile *ctrl = &pq->bar0[BAR0_CTRL];

	/* Removed CTRL_CLR_ZERO, as we need it for unframed mode. */
	*ctrl =
	    CTRL_ALTRES | CTRL_SDIOD1 | CTRL_SDIOD2 |
	    CTRL_MASL | CTRL_CT_TI_EN | CTRL_CONS_DEL;
	udelay(1000);
	*ctrl =
	    CTRL_SDIOD1 | CTRL_SDIOD2 |
	    CTRL_MASL | CTRL_CT_TI_EN | CTRL_CONS_DEL |
	    CTRL_DMARST | CTRL_TMS_DF;

	udelay(1000);
	pq->bar0[BAR0_IMR] = 0x80000000;
	pq->bar0[BAR0_BSLS] = 0x00000003; // BS/LS: allow to sync to all E1 channels

	//initialize dma
    pq->bar0[BAR0_DMA_ADDR] = pq->readdma;
    pq->bar0[BAR0_DMA_RADDR] = pq->writedma;
    pq->bar0[BAR0_DMACR] = DMA_IFQ2 | DMA_SIZE0;
}

static void init_core_regs_quasar8(struct quasar * pq)
{
	unsigned int volatile *ctrl = &pq->bar0[BAR0_CTRL];
    
    /* Configure board */
    *ctrl = 
        CTRL_ALTRES | CTRL_SDIOD1 | CTRL_MASL | CTRL_CT_TI_EN |
        CTRL_DMARST | CTRL_CONS_DEL;
	
    if (PMC_OLD_FIRMWARE != (pq->bar0[BAR0_HW_FW_REV] & 0xffff)) 
        *ctrl |= CTRL_NRFREQ;

    udelay(1000);

    /* Wake up framers, hm magic reset */
    *ctrl = 
        CTRL_SDIOD1 | CTRL_MASL | CTRL_CT_TI_EN |
        CTRL_DMARST | CTRL_CONS_DEL |
        CTRL_PTENOVR | CTRL_MTRESET | CTRL_FLOCK;
	
    udelay(1000);
    pq->bar0[BAR0_IMR] = 0x80000000;	/* Enable interrupts, in general */
    pq->bar0[BAR0_BSLS] = 0x000000ff;	/* BS/LS: allow to sync to all E1 channels */
    pq->bar0[BAR0_TXAO] = 0x000000ff;	/* TX_enable: enable transmit */


   //initialize dma
    pq->bar0[BAR0_DMA_ADDR] = pq->readdma;
    pq->bar0[BAR0_DMA_RADDR] = pq->writedma;
        pq->bar0[BAR0_DMACR] = DMA_IFQ0 | DMA_SIZE0;
}

static void init_core_regs_quasar2012(struct quasar * pq)
{
	unsigned int volatile *ctrl = &pq->bar0[BAR0_CTRL];

	*ctrl= 
	    CTRL_ALTRES | CTRL_SDIOD1 | CTRL_SDIOD2 | 
	    CTRL_CT_TI_EN | CTRL_CONS_DEL;
	udelay(1000);

	*ctrl=
	    CTRL_SDIOD1 | 
	    CTRL_CT_TI_EN | CTRL_CONS_DEL | 
	    CTRL_DMARST;
	udelay(1000);

	pq->bar0[BAR0_IMR] = 0x80000000; // enable interrupts on PCI
	pq->bar0[BAR0_BSLS] = 0x000000ff; // all ports are  masters (freerun)

   //initialize dma
    pq->bar0[BAR0_DMA_ADDR] = pq->readdma;
    pq->bar0[BAR0_DMA_RADDR] = pq->writedma;
    pq->bar0[BAR0_DMACR] = DMA_IFQ0 | DMA_SIZE0;

}

static void init_core_regs_quasar16(struct quasar * pq)
{
	unsigned int volatile *ctrl = &pq->bar0[BAR0_CTRL];

	*ctrl= 
	    CTRL_ALTRES | CTRL_SDIOD1 | CTRL_SDIOD2;
	udelay(1000);

	*ctrl=
	    CTRL_SDIOD1 | CTRL_DMARST;
	udelay(1000);

	pq->bar0[BAR0_IMR] = 0x80000000; // enable interrupts on PCI
	pq->bar0[BAR0_BSLS] = 0x0000ffff; // all ports are  masters (freerun)

       //initialize dma
        pq->bar0[BAR0_DMA_ADDR] = pq->readdma;

    // DMA size 8 K = 16 ports * 32 ts * 8 chunk * 2 buffers
    // interrupts 1/2 buffer size
   	pq->bar0[BAR0_DMACR] = DMA_IFQ0 | DMA_SIZE1;

}
static void init_switch_map_quasar16(struct quasar *pq)
{
	// using hardware preset
	return;
}


static void init_switch_map_quasar2012(struct quasar *pq)
{
	int i, j;

	/*
	   PC=>E1
	   DMA_LINE_2 => CT_BUS_1 => E1_LINE0
	   DMA_LINE_3 => CT_BUS_3 => E1_LINE1 */

	for (i = 0; i < 128; i++) {	/* LINE2,3 -> CT_BUS1,3 */
		pq->bar0[0x2100 + i] = E1_TO_CTB_TABLE(2, i);
		pq->bar0[0x2300 + i] = E1_TO_CTB_TABLE(3, i);
	}
	for (i = 0; i < 32; i++) {
		/* CT_BUS => E1; according to suggestion 
		   4 CT channel
		   CT_BUS1 -> LINE0 */
		pq->bar0[0x1000 + i] = CTB_TO_E1_TABLE(1, i);
		pq->bar0[0x1100 + i] = CTB_TO_E1_TABLE(1, (i + 32));
		pq->bar0[0x1200 + i] = CTB_TO_E1_TABLE(1, (i + 64));
		pq->bar0[0x1300 + i] = CTB_TO_E1_TABLE(1, (i + 96));
		/* CT_BUS3 -> LINE1 */
		pq->bar0[0x1400 + i] = CTB_TO_E1_TABLE(3, i);
		pq->bar0[0x1500 + i] = CTB_TO_E1_TABLE(3, (i + 32));
		pq->bar0[0x1600 + i] = CTB_TO_E1_TABLE(3, (i + 64));
		pq->bar0[0x1700 + i] = CTB_TO_E1_TABLE(3, (i + 96));
	}
	/* E1=>PC */
	for (i = 0; i < 32; i++) {
		/* E1 => CT_BUS */
		for (j = 0; j < 4; j++) {
			/* E1_LINE0 => CT_BUS_0 */
			pq->bar0[0x2000 + i + 32 * j] =
			    E1_TO_CTB_TABLE(0, (i * 4 + j));
			/* E1_LINE1 => CT_BUS_2 */
			pq->bar0[0x2200 + i + 32 * j] =
			    E1_TO_CTB_TABLE(1, (i * 4 + j));
		}
	}

	if (pq->ec) pq->bar0[BAR0_DMA_CT_CFG] = 0x64;
        else pq->bar0[BAR0_DMA_CT_CFG] = 0x20;

	pq->bar0[BAR0_CTRL] |= CTRL_DSP_RST;	// reset DSP
}

static void init_switch_map_quasar8(struct quasar *pq)
{
	int i, j;

	/*
	   PC=>E1
	   DMA_LINE_2 => CT_BUS_4 => E1_LINE0
	   DMA_LINE_3 => CT_BUS_5 => E1_LINE1 */

	for (i = 0; i < 128; i++) {	/* LINE2 -> CT_BUS4 */
		pq->bar0[0x2400 + i] = E1_TO_CTB_TABLE(2, i);
		pq->bar0[0x2500 + i] = E1_TO_CTB_TABLE(3, i);
	}
	for (i = 0; i < 32; i++) {
		/* CT_BUS => E1; according to suggestion 
		   4 CT channel
		   CT_BUS4 -> LINE0 */
		pq->bar0[0x1000 + i] = CTB_TO_E1_TABLE(4, i);
		pq->bar0[0x1100 + i] = CTB_TO_E1_TABLE(4, (i + 32));
		pq->bar0[0x1200 + i] = CTB_TO_E1_TABLE(4, (i + 64));
		pq->bar0[0x1300 + i] = CTB_TO_E1_TABLE(4, (i + 96));
		/* CT_BUS5 -> LINE1 */
		pq->bar0[0x1400 + i] = CTB_TO_E1_TABLE(5, i);
		pq->bar0[0x1500 + i] = CTB_TO_E1_TABLE(5, (i + 32));
		pq->bar0[0x1600 + i] = CTB_TO_E1_TABLE(5, (i + 64));
		pq->bar0[0x1700 + i] = CTB_TO_E1_TABLE(5, (i + 96));
	}
	/* E1=>PC */
	for (i = 0; i < 32; i++) {
		/* E1 => CT_BUS */
		for (j = 0; j < 4; j++) {
			/* E1_LINE0 => CT_BUS_6 => WRITE_BUF_0 */
			pq->bar0[0x2600 + i + 32 * j] =
			    E1_TO_CTB_TABLE(0, (i * 4 + j));
			/* E1_LINE1 => CT_BUS_7 => WRITE_BUF_1 */
			pq->bar0[0x2700 + i + 32 * j] =
			    E1_TO_CTB_TABLE(1, (i * 4 + j));
		}
	}
	i = 0x76;		/* 6CT=> CHAN0 7CT=>CHAN1 */
	pq->bar0[BAR0_DMA_CT_CFG] = i;
	/* Mapping to DMA buffers is the same as in case of acess point first come 0 framer and 0 time slot */
}

static void init_switch_map_quasarm(struct quasar * pq)
{
    int i;

    if (pq->ec) {
	for (i=0;i<16;i++){
	//LINE0(E1_0 tx) <- LINE2(DMA READ BUFFER(quasar->pc))
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1000+i*2] = INPUT_TS_SIGNATURE(2, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1000+i*2+1] = INPUT_TS_SIGNATURE(2, (i*2+1));

	//LINE1(E1_0 tx) <- LINE3(DMA READ BUFFER(quasar->pc))
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1100+i*2] = INPUT_TS_SIGNATURE(3, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1100+i*2+1] = INPUT_TS_SIGNATURE(3, (i*2+1));

	// LINE 2 (DMA WRITE BUF) <- LINE4,6 (DSP DX0.0..31)
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1200+i*2] = INPUT_TS_SIGNATURE(4, i);
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1200+i*2+1] = INPUT_TS_SIGNATURE(6, i);

	// LINE 3 (DMA WRITE BUF) <- LINE4,6 (DSP DX0.32..63)
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1300+i*2] = INPUT_TS_SIGNATURE(4, (i+16));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1300+i*2 + 1] = INPUT_TS_SIGNATURE(6, (i+16));

	// LINE 4,6 (DSP DR0.0..31) <- E1_0 rx
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1400+i] = INPUT_TS_SIGNATURE(0, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1600+i] = INPUT_TS_SIGNATURE(0, (i*2+1));

	// LINE 4,6 (DSP DR0.32..63) <- E1_1 rx
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1400+i+16] = 	INPUT_TS_SIGNATURE(1, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1600+i+16] = 	INPUT_TS_SIGNATURE(1, (i*2+1));

	// LINE 5,7 (DSP DR1.0..31) <- DMA READ BUFFER
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1500+i] = INPUT_TS_SIGNATURE(2, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1700+i] = INPUT_TS_SIGNATURE(2, (i*2+1));

	// LINE 5,7 (DSP DR1.32..63) <- DMA READ BUFFER
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1500+i+16] = 	INPUT_TS_SIGNATURE(3, (i*2));
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1700+i+16] = 	INPUT_TS_SIGNATURE(3, (i*2+1));
	}
	
	// reset DSP
        pq->bar0[BAR0_CTRL] |= CTRL_DSP_RST;
    }
    else {
//Mapping of the timeslots for the Quasar Mini
//////////From E1 to PC///////////////////////////////////////////////
    //LINE2(DMA WRITE BUFFER(quasar->pc)) <- LINE0(E1_0 input)
	for(i=0;i<32;i++) {
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1200+i] = INPUT_TS_SIGNATURE(0, i);
	}
    //LINE3(DMA WRITE BUFFER(quasar->pc)) <- LINE1(E1_1 input)
	for(i=0;i<32;i++) {
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1300+i] = INPUT_TS_SIGNATURE(1, i);
        }
/////////From PC to E1////////////////////////////////////////////////
    //LINE0(E1_0 output) <- LINE2(DMA READ BUFFER(quasar->pc))
	for(i=0;i<32;i++) {
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1000+i] = INPUT_TS_SIGNATURE(2, i);
        }
    //LINE1(E1_1 output) <- LINE3(DMA READ BUFFER(quasar->pc))
	for(i=0;i<32;i++) {
	    pq->bar0[SWITCH_TABLE_OFFSET + 0x1100+i] = INPUT_TS_SIGNATURE(3, i);
        }
    }
}

static void set_hw_options(struct quasar * pq, int noec)
{
	struct pci_dev *pdev = pq->pci_device;
	u8 rev_id=0;

	pci_read_config_byte(pdev, PCI_REVISION_ID, &rev_id);

	if (QUASARM(pdev) || QUASARMICRO(pdev))
		pq->e1_amount = 1; 	/* 2 channels by default */
	if (QUASAR8(pdev) || QUASAR2012(pdev))
		pq->e1_amount = 7; 	/* 8 channels by default */
	if (QUASAR16(pdev))
	    pq->e1_amount = 15; /* 16 channels by default */

	/* Read channels number, if specified. */
	if (0xa0 == (rev_id & 0xf0))
    	pq->e1_amount = (rev_id & 0x0f) - 1;

	if ((pdev->subsystem_device & SUBSYSTEM_EC) && (noec == 0)) 
	    pq->ec = 1;

	if (pdev->subsystem_device & SUBSYSTEM_WD)
	    pq->wd = 1;
}


static void init_framer_regs_quasar8(struct quasar * pq)
{
	int addr = 0x800, j;

	if (PMC_OLD_FIRMWARE == (pq->bar0[BAR0_HW_FW_REV] & 0xffff)) {
		pq->bar0[0x03] = 0x00;	/* reset all framers */
		udelay(1000);
		pq->bar0[0x03] = 0xff;
		udelay(1000);
	}

	pq->bar1[addr] = 0x02;	/* setup page 2 */
	pq->bar1[addr + 0x10] = 0xa0;	/* Configuration control word: E1 mode, Transmit enable */
	pq->bar1[addr + 0x11] = 0;	/* P2.11=0h (Pulse attenuation) for 120 Om line. LIU control word */
	pq->bar1[addr + 0x12] = 0;	/* write 0's to reserved space */
	pq->bar1[addr + 0x14] = 0;
	pq->bar1[addr + 0x15] = 0;
	pq->bar1[addr + 0x16] = 0xbb;	/* P2.16 eqalizer high threshold */
	pq->bar1[addr + 0x17] = 0x30;	/* eqalizer low threshold */

	pq->bar1[addr] = 0x01;	/* setup page 1 */
	pq->bar1[addr + 0x14] = 0x80;	/* to output data to dsto, select CAS */
	pq->bar1[addr + 0x19] = 0x01;	/* to obtain loss signal */
	pq->bar1[addr + 0x1f] = 0x18;	/* to setup automatic equaliser */
	pq->bar1[addr + 0x15] = 0xa0;	/* P1.15 Coding and loopback controll HDB both directions */
	pq->bar1[addr + 0x13] = 0x0b;	/* setup defaults for x bits? X123=111b */
	pq->bar1[addr + 0x12] = 0x03;	/* setup defaults for International bits I01=11b */
	pq->bar1[addr + 0x11] = 0x10;	/* disable TxAO */

	/* cas related */
	pq->bar1[addr + 0x1e] = 0x1;	/* enable CAS interrupt */

	/* Enable to transmit CAS from registers on 5&6 pages */
	/* setup page 7 */
	pq->bar1[addr] = 0x07;

	for (j = 0; j < 16; j++) {
		pq->bar1[addr + 0x10 + j] = 0x02;

	}
	pq->bar1[addr] = 0x08;
	for (j = 0; j < 16; j++) {
		pq->bar1[addr + 0x10 + j] = 0x02;
	}
	/* fill 15ts CAS DATA nible with something to avoid starange initial shiftion of CAS, looks like framer hardware bug */
	pq->bar1[addr] = 0x5;
	pq->bar1[addr + 0x1f] = 0x5;
}
int acc_lxt_reg(struct quasar * pq ,unsigned char chan_num, unsigned char addr, unsigned char mode, unsigned char *data) {

    if (chan_num > 1) return -1;
    if (mode > 1) return -1;
    if (!pq) return -1;
    if (addr > 7) return -1;


    pq->bar0[BAR0_LIU_CTRL] = 0;
    pq->bar0[BAR0_LIU_CTRL] = (chan_num << 10) | LIU_CS; //set cs for definite channel;
    addr = (addr | 0x10 ) << 1;
    pq->bar0[BAR0_LIU_CTRL] = (chan_num << 10) | LIU_CS | LIU_TR | addr | mode;  //pass address and operation
    while (pq->bar0[BAR0_LIU_CTRL] & LIU_TR) {};
    if (LIU_WRITE == mode) {
        pq->bar0[BAR0_LIU_CTRL] = ((chan_num) << 10) | LIU_CS | LIU_TR | *data;
        while (pq->bar0[BAR0_LIU_CTRL] & LIU_TR) {};
    } else {
        pq->bar0[BAR0_LIU_CTRL] = ((chan_num) << 10) | LIU_CS | LIU_TR;
        while (pq->bar0[BAR0_LIU_CTRL] & LIU_TR) {};
        *data = 0xff & pq->bar0[BAR0_LIU_CTRL];
    }
    pq->bar0[BAR0_LIU_CTRL] = 0;
    return 0;
}

static void init_framer_regs_quasar16(struct quasar *pq)
{
	unsigned char data, i;
	unsigned int FRC1;

	FRC1 = FRC_ARAI | FRC_EBIT_TX1 | FRC_EBIT_TX0 | FRC_FR_RX_EN | FRC_FR_TX_EN | FRC_CRC4_TX_EN | FRC_MFR_TX_EN;

	for (i = 0; i <= pq->e1_amount; i++) {
		module_printk(KERN_INFO, "%s: Using short haul mode on port %d\n",
				      pq->name, i);
		pq->have_liu[i] = 0;
		pq->bar1[(0x40 * i) + FRC1_ADDR_OFFSET] = FRC1 | FRC_SHORT_HAUL_EN |
			FRC_TXEN | FRC_HDB3_RX_EN | FRC_HDB3_TX_EN;

		module_dprintk(KERN_DEBUG, "Framer config register: 0x%x\n",
			       pq->bar1[(0x40 * i)]);
		pq->bar1[(0x40 * i) + CAS_BIT_BUFFER_OFFSET] = 0x0b;
		/* Set transmit 5 on abcd bits for all available channels */
		for (data = 1; data < 16; data++) {
			pq->bar1[(0x40 * i) + CAS_BIT_BUFFER_OFFSET + data] = 0x55;
		}
	}
}

static void init_framer_regs_quasarm(struct quasar *pq)
{
	unsigned char data, i;
	unsigned int FRC1;

	FRC1 = FRC_ARAI | FRC_EBIT_TX1 | FRC_EBIT_TX0 | FRC_FR_RX_EN | FRC_FR_TX_EN | FRC_CRC4_TX_EN | FRC_MFR_TX_EN;

	for (i = 0; i <= pq->e1_amount; i++) {
		data = 1;
		acc_lxt_reg(pq, i, LR_CR1, LIU_WRITE, &data);
		data = 0;
		acc_lxt_reg(pq, i, LR_CR1, LIU_READ, &data);

		if (1 == data) {
			module_printk(KERN_INFO, "%s: LIU found on port %d\n", pq->name, i);
			pq->have_liu[i] = 1;
			pq->bar1[(0x40 * i) + FRC1_ADDR_OFFSET] = FRC1;
			data = LR_CR2_RESET;
			acc_lxt_reg(pq, i, LR_CR2, LIU_WRITE, &data);
			mdelay(35);
			data = LR_CR2_ENLOOP;	/* enables NLOOP detection */
			acc_lxt_reg(pq, i, LR_CR2, LIU_WRITE, &data);
			/* HDB3, long haul, JA-tx */
			data = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4;
			acc_lxt_reg(pq, i, LR_CR1, LIU_WRITE, &data);
			data = 0;
			acc_lxt_reg(pq, i, LR_CR1, LIU_READ, &data);
			module_dprintk(KERN_DEBUG, "LIU CR1 register = 0x%02x\n", data);
			data = LR_CR4_CODEV | LR_CR4_ZEROV;
			acc_lxt_reg(pq, i, LR_CR4, LIU_WRITE, &data);
		} else {
			module_printk(KERN_INFO, "%s: Using short haul mode on port %d\n",
				      pq->name, i);
			pq->have_liu[i] = 0;
			pq->bar1[(0x40 * i) + FRC1_ADDR_OFFSET] = FRC1 | FRC_SHORT_HAUL_EN |
				FRC_TXEN | FRC_HDB3_RX_EN | FRC_HDB3_TX_EN;
		}
		module_dprintk(KERN_DEBUG, "Framer config register: 0x%x\n",
			       pq->bar1[(0x40 * i)]);
		pq->bar1[(0x40 * i) + CAS_BIT_BUFFER_OFFSET] = 0x0b;
		/* Set transmit 5 on abcd bits for all available channels */
		for (data = 1; data < 16; data++) {
			pq->bar1[(0x40 * i) + CAS_BIT_BUFFER_OFFSET + data] = 0x55;
		}
	}
}

/*========== IOCTL (ecfg) ==========*/
/*
 static void check_bit(..)
 to_hw - structure to pass configuration
 from_hw - structure to return actual settings from hardware back
 smask - position of property bit in configuration structure
 hmask - position of same bit in hardware register
*/

static void check_bit_m(struct quasar * pq,
			unsigned char chan_num, unsigned char smask,
			unsigned int hmask, unsigned char clear_hwbit_if_swbit_one,
			unsigned char to_hw, unsigned char * from_hw)
{
    unsigned int tmp;
    tmp = pq->bar1[0x40*chan_num];
    tmp &= ~hmask; //clear bit;
    if(to_hw & smask){
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            tmp |= hmask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            tmp |= hmask;
        } else {

        }
    }
    pq->bar1[0x40*chan_num] = tmp;
    //read back
    tmp = 0;
    tmp = pq->bar1[0x40*chan_num];
    if(tmp & hmask)
    {
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            *from_hw |= smask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            *from_hw |= smask;
        } else {
            
        }
    }
    return;
}

static void align_to_swposition_m(struct quasar * pq, unsigned char chan_num, unsigned char smask,
                                unsigned int hmask, unsigned char clear_hwbit_if_swbit_one, unsigned char * from_hw) {
    unsigned int tmp=0;
    tmp = pq->bar1[0x40*chan_num];
    if(tmp & hmask)
    {
        if (clear_hwbit_if_swbit_one){

        } else { //set bit
            *from_hw |= smask;
        }

    } else {
        if (clear_hwbit_if_swbit_one){
            *from_hw |= smask;
        } else {
            
        }
    }
    return;
}

/*
 Returns:
 ENOTTY - in case of invalid ioctl;
 EINVAL - in case of passing invalid parameters;
 EFAULT - in case of error in the data transmitin beetween user and kernel space;
 ENOSYS - Function not implemented
 0 - in case of success;
*/

static int ioctl_quasarm(struct inode * inode_p, struct file * file_p, unsigned int cmd, unsigned long arg)
{
	unsigned int tmp=0, j;
    unsigned char ucTmp=0;
	unsigned long ulTmp=0;
	struct pmc_config_chan store_chanl_conf, read_back_chan_conf;
	struct pmc_err_ins err_ins;
	struct pmc_data_pass pass_data;
	struct pmc_status_chan stat_exchange;
	struct pmc_config_board board_config;
        struct pmc_cas_xmit_data cas_data;
	struct pmc_version version_exchange;
	struct quasar * pq = (struct quasar *) file_p->private_data;

	memset((void *) &board_config, 0, sizeof(struct pmc_config_board));
	memset((void *) &read_back_chan_conf, 0, sizeof(struct pmc_config_chan));
	memset((void *) &store_chanl_conf, 0, sizeof(struct pmc_config_chan));
        memset((void *) &err_ins, 0, sizeof(struct pmc_err_ins));
	memset((void *) &stat_exchange, 0, sizeof(struct pmc_status_chan));
        memset((void *) &cas_data, 0, sizeof(struct pmc_cas_xmit_data));
	if (_IOC_TYPE(cmd) != PTMC_E1_8_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > PTMC_E1_8_MCN)
		return -ENOTTY;

	switch(cmd) {
	case PTMC_E1_CHAN_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CHAN_CONFIGURE.\n");
		if(copy_from_user((void *) &store_chanl_conf, (void *) arg, sizeof(store_chanl_conf)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if ((store_chanl_conf.chan_num & (~DIRECT_IN)) > pq->e1_amount) return -EINVAL;

		module_dprintk(KERN_DEBUG, "Before config, chan:%d data: %x\n", store_chanl_conf.chan_num, pq->bar1[0x40 * store_chanl_conf.chan_num]);

		if (!(store_chanl_conf.chan_num & DIRECT_IN)) //writing cas data
		{
			store_chanl_conf.chan_num&= (~DIRECT_IN);

			//cfg0

			//pq, chan_num,                           , smask   ,     hmask,         Nbit, to_hw                ,  from_hw
			check_bit_m(pq, store_chanl_conf.chan_num, CFG0_CAS,     FRC_MFR_TX_EN, 1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );
			check_bit_m(pq, store_chanl_conf.chan_num, CFG0_RCV_TR,  FRC_FR_RX_EN,  1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );
			check_bit_m(pq, store_chanl_conf.chan_num, CFG0_XMIT_TR, FRC_FR_TX_EN , 1   , store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0 );

			tmp = pq->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				check_bit_m(pq, store_chanl_conf.chan_num, CFG0_RCV_HDB3, FRC_HDB3_RX_EN, 0, store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
				check_bit_m(pq, store_chanl_conf.chan_num, CFG0_XMIT_HDB3, FRC_HDB3_TX_EN, 0, store_chanl_conf.cfg0,  &read_back_chan_conf.cfg0);
			} else {
				module_dprintk(KERN_DEBUG, "Configure HDB3 on long haul port\n"); //TODO,  ! on framer only both tx + rx
				if((store_chanl_conf.cfg0 & CFG0_RCV_HDB3) || (store_chanl_conf.cfg0 & CFG0_XMIT_HDB3)) { //if HDB3 on RX or TX enable HDB3 on LIU
					ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4; //HDB3, long haul, JA-tx;
					module_dprintk(KERN_DEBUG, "To write ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR1, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
					module_dprintk(KERN_DEBUG, "Read back ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					if (ucTmp & LR_CR1_ENCENB) {
						read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
						read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
					} else {
						module_printk(KERN_ERR, "LIU configuration error\n");
					}
				} else {
					module_dprintk(KERN_DEBUG, "Configure AMI on long haul port\n"); //TODO,  ! on framer only both tx + rx
					ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_EC1 | LR_CR1_EC4;
					module_dprintk(KERN_DEBUG, "To write ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR1, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
					module_dprintk(KERN_DEBUG, "Read back ucTmp =0x%02x ch_num=%d\n", ucTmp, store_chanl_conf.chan_num);
					if (ucTmp & LR_CR1_ENCENB)
					{
						read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
						read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
						module_printk(KERN_ERR, "LIU configuration error\n");
					}

				}
			}

			//configure board part
			if (store_chanl_conf.cfg0 & CFG0_MASTER)
			{
				tmp = pq->bar0[BAR0_BSLS]; //get bs/ls
				ucTmp = (1<<store_chanl_conf.chan_num);
				tmp|= (unsigned int) ucTmp;
				pq->bar0[BAR0_BSLS] = tmp; //restore bs/ls
				read_back_chan_conf.cfg0|= CFG0_MASTER;
			}
			else                                          //allow to sync to this chan
			{
				tmp = pq->bar0[BAR0_BSLS]; //get bs/ls
				ucTmp = ~(1<<store_chanl_conf.chan_num);
				tmp&= (unsigned int) ucTmp;
				pq->bar0[BAR0_BSLS] = tmp; //restore bs/ls
			}



			//CFG1
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_NAT0, FRC_SA_BITS0, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_NAT1, FRC_SA_BITS1, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_NAT2, FRC_SA_BITS2, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_NAT3, FRC_SA_BITS3, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_NAT4, FRC_SA_BITS4, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_INAT0, FRC_EBIT_TX0, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG1_INAT1, FRC_EBIT_TX1, 0, store_chanl_conf.cfg1,  &read_back_chan_conf.cfg1);


			//CFG2
			check_bit_m(pq, store_chanl_conf.chan_num, CFG2_RAI_MANUAL, FRC_ARAI, 0, store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG2_RAI_ON, FRC_RAI_TX, 0, store_chanl_conf.cfg2,  &read_back_chan_conf.cfg2);
			//CFG2_RCAS_AUTO_ON //not implemented

			//CFG2_RCAS_ON
			if (store_chanl_conf.cfg2 & CFG2_RCAS_ON) { //enable MF alarm (Y -bit -> 1)
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_Y_BIT;
			} else {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_Y_BIT;
			}
			//read back
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_Y_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_RCAS_ON;
			}

			//CFG2_X1
			if (store_chanl_conf.cfg2 & CFG2_X1) {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X1_BIT;
			} else {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X1_BIT;
			}
			//read back
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X1_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X1;
			}
			//CFG2_X2
			if (store_chanl_conf.cfg2 & CFG2_X2) {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X2_BIT;
			} else {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X2_BIT;
			}
			//read back
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X2_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X2;
			}
			//CFG2_X3
			if (store_chanl_conf.cfg2 & CFG2_X3) {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] |= CAS_X3_BIT;
			} else {
				pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] &= ~CAS_X3_BIT;
			}
			//read back
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X3_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X3;
			}

			//CFG3

			//CFG3_CRC4_TR ignore
			check_bit_m(pq, store_chanl_conf.chan_num, CFG3_CRC4_DIS, FRC_CRC4_TX_EN, 1, store_chanl_conf.cfg3,  &read_back_chan_conf.cfg3);
			//CFG3_CRC4_AUTO ignore

			//CFG4
			check_bit_m(pq, store_chanl_conf.chan_num, CFG4_AIS,   FRC_TXAO, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			check_bit_m(pq, store_chanl_conf.chan_num, CFG4_RLOOP, FRC_RLBK, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			//check_bit_m(pq, store_chanl_conf.chan_num, CFG4_PLOOP, FRC_, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);

			tmp = pq->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				check_bit_m(pq, store_chanl_conf.chan_num, CFG4_LLOOP, FRC_MLBK, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
				check_bit_m(pq, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			} else {
				if(store_chanl_conf.cfg4 & CFG4_LLOOP) {
					module_dprintk(KERN_DEBUG, "Enable local loopback on long haul port\n");
					ucTmp =	0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					ucTmp |= LR_CR2_EALOOP;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					if (ucTmp & LR_CR2_EALOOP) {
						read_back_chan_conf.cfg4 |= CFG4_LLOOP;
					} else {
						module_printk(KERN_ERR, "LIU configuration error\n");
					}
				} else {
					module_dprintk(KERN_DEBUG, "Disable local loopback on long haul port\n");
					ucTmp =	0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					ucTmp &= ~LR_CR2_EALOOP;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_WRITE, &ucTmp);
					ucTmp = 0;
					acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
					if (ucTmp & LR_CR2_EALOOP) {
						module_printk(KERN_ERR, "LIU configuration error\n");
						read_back_chan_conf.cfg4 |= CFG4_LLOOP;
					}
				}
				//	check_bit_m(pq, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);
			}



		}
		else //only read data;
		{
			store_chanl_conf.chan_num&= (~DIRECT_IN);


			//cfg0
			//pq, chan_num,                        , smask   ,     hmask,       , Nbit,  from_hw
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG0_CAS,     FRC_MFR_TX_EN, 1   , &read_back_chan_conf.cfg0 );
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG0_RCV_TR,  FRC_FR_RX_EN,  1   , &read_back_chan_conf.cfg0 );
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG0_XMIT_TR, FRC_FR_TX_EN , 1   , &read_back_chan_conf.cfg0 );
			tmp = pq->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG0_RCV_HDB3, FRC_HDB3_RX_EN, 0,  &read_back_chan_conf.cfg0);
				align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG0_XMIT_HDB3, FRC_HDB3_TX_EN, 0, &read_back_chan_conf.cfg0);
			} else {
				ucTmp = 0;
				acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR1, LIU_READ, &ucTmp);
				if (ucTmp & LR_CR1_ENCENB) {
					read_back_chan_conf.cfg0 |= CFG0_RCV_HDB3;
					read_back_chan_conf.cfg0 |= CFG0_XMIT_HDB3;
				}
			}
			tmp = pq->bar0[BAR0_BSLS]; //get bs/ls
			tmp = tmp >> store_chanl_conf.chan_num;
			if (tmp & 1) {
				read_back_chan_conf.cfg0 |= CFG0_MASTER;
			}

			//cfg1
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_NAT0, FRC_SA_BITS0, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_NAT1, FRC_SA_BITS1, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_NAT2, FRC_SA_BITS2, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_NAT3, FRC_SA_BITS3, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_NAT4, FRC_SA_BITS4, 0,        &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_INAT0, FRC_EBIT_TX0, 0,       &read_back_chan_conf.cfg1);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG1_INAT1, FRC_EBIT_TX1, 0,       &read_back_chan_conf.cfg1);

			//cfg2
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG2_RAI_MANUAL, FRC_ARAI, 0,     &read_back_chan_conf.cfg2);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG2_RAI_ON, FRC_RAI_TX, 0,       &read_back_chan_conf.cfg2);

			//CFG2_RCAS_ENA
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_Y_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_RCAS_ON;
			}
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X3_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X3;
			}
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X2_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X2;
			}
			if (pq->bar1[(0x40 * store_chanl_conf.chan_num) + CAS_BIT_BUFFER_OFFSET] & CAS_X1_BIT) {
				read_back_chan_conf.cfg2 |= CFG2_X1;
			}

			//CFG3

			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG3_CRC4_DIS, FRC_CRC4_TX_EN, 1,   &read_back_chan_conf.cfg3);


			//cfg4
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG4_AIS   , FRC_TXAO, 	      1,   &read_back_chan_conf.cfg4);
			align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG4_RLOOP , FRC_RLBK,        0,   &read_back_chan_conf.cfg4);
			//    	check_bit_m(pq, store_chanl_conf.chan_num, CFG4_PLOOP, FRC_, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4);

			tmp = pq->bar1[0x40 * store_chanl_conf.chan_num];

			if (tmp & FRC_SHORT_HAUL_EN) { //short haul
				align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG4_LLOOP, FRC_MLBK, 0,  &read_back_chan_conf.cfg4);
				align_to_swposition_m(pq, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXEN, 1, &read_back_chan_conf.cfg4);
			} else {
				ucTmp = 0;
				acc_lxt_reg(pq, store_chanl_conf.chan_num, LR_CR2, LIU_READ, &ucTmp);
				if (ucTmp & LR_CR2_EALOOP) read_back_chan_conf.cfg4 |= CFG4_LLOOP;
				//check_bit_m(pq, store_chanl_conf.chan_num, CFG4_TX_DIS, FRC_TXAO, 0, store_chanl_conf.cfg4,  &read_back_chan_conf.cfg4); //no equvivalent in LIU Mode
			}

		}
		module_dprintk(KERN_DEBUG, "After config, chan: %d data: %x\n", store_chanl_conf.chan_num, pq->bar1[0x40 * store_chanl_conf.chan_num]);
		if(copy_to_user((void *) arg, (void *) &read_back_chan_conf, sizeof(read_back_chan_conf)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_ERR_EMU:
		module_printk(KERN_WARNING, "PTMC_E1_ERR_EMU.\n");
		module_printk(KERN_WARNING, "Unimplemented.\n");
		return -ENOSYS;
		//AIS

		break;
	case PTMC_E1_GET_STATS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_STATS.\n");
		if(copy_from_user((void *) &stat_exchange, (void *) arg, sizeof(stat_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (stat_exchange.chan_num > pq->e1_amount) return -EINVAL;

		//stat 0
		tmp = pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_ADDR_OFFSET];
		stat_exchange.stat0 = 0;

		if (tmp & FRS_LOS) stat_exchange.stat0 |= STAT0_LOS;
		if (tmp & FRS_RED) stat_exchange.stat0 |= STAT0_LOF;
		if (!(tmp & FRS_MFRAME_SYNC)) stat_exchange.stat0 |= STAT0_CAS_LOM;
		if (!(tmp & FRS_CRC_SYNC)) stat_exchange.stat0 |= STAT0_CRC4_SYNC;
		if (!((tmp & FRS_EBIT_RX1) && (tmp & FRS_EBIT_RX0))) stat_exchange.stat0 |= STAT0_E_BIT;
		if (tmp & FRS_CRC4_ERR) stat_exchange.stat0 |= STAT0_CRC4;
		if (tmp & FRS_AIS) stat_exchange.stat0 |= STAT0_AIS;
		if (tmp & FRS_RAI_RX) stat_exchange.stat0 |= STAT0_RAIS;

		//stat1
		stat_exchange.stat1 = 0;
		if (tmp & FRS_SA_BITS0) stat_exchange.stat1 |= STAT1_NAT0;
		if (tmp & FRS_SA_BITS1) stat_exchange.stat1 |= STAT1_NAT1;
		if (tmp & FRS_SA_BITS2) stat_exchange.stat1 |= STAT1_NAT2;
		if (tmp & FRS_SA_BITS3) stat_exchange.stat1 |= STAT1_NAT3;
		if (tmp & FRS_SA_BITS4) stat_exchange.stat1 |= STAT1_NAT4;
		if (tmp & FRS_EBIT_RX0) stat_exchange.stat1 |= STAT1_INAT0;
		if (tmp & FRS_EBIT_RX1) stat_exchange.stat1 |= STAT1_INAT1;
		if (tmp & FRS_SLIP_ERR) stat_exchange.stat1 |= STAT1_SLIP;

		//stat2
		stat_exchange.stat2 = 0;
		if (tmp & FRS_MFRAI) stat_exchange.stat2 |= STAT2_Y;
		tmp = pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET];
		if (tmp & CAS_RX_X1_BIT) stat_exchange.stat2 |= STAT2_X1;
		if (tmp & CAS_RX_X2_BIT) stat_exchange.stat2 |= STAT2_X2;
		if (tmp & CAS_RX_X3_BIT) stat_exchange.stat2 |= STAT2_X3;

		//hdb3
		tmp = pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_HDB3ERR_OFFSET];
		stat_exchange.hdb3 = (unsigned short) tmp & 0xFFFF;

		//crc4
		tmp = pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_CRC4ERR_OFFSET];
		stat_exchange.crc4 = (unsigned short) tmp & 0xFFFF;

		//fas
		tmp = pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_FRAMEERR_OFFSET];
		stat_exchange.fas = (unsigned short) tmp & 0xFFFF;

		//mgmt

		if (stat_exchange.mgmt & 0x01) pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_HDB3ERR_OFFSET] = 0;
		if (stat_exchange.mgmt & 0x02) pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_FRAMEERR_OFFSET] = 0;
		if (stat_exchange.mgmt & 0x04) pq->bar1[(0x40 * stat_exchange.chan_num) + FRS_CRC4ERR_OFFSET] = 0;

		//ovr
		stat_exchange.ovr = 0;

		//get cas signaling

		for (j=0; j<16; j=j+2) {
			tmp = (pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 8) & 0xf0;
			stat_exchange.cas_data[j/2] = (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=1; j<16; j=j+2) {
			tmp = (pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 12) & 0xf;
			stat_exchange.cas_data[j/2] |= (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=0; j<16; j=j+2) {
			tmp = (pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 4) & 0xf0;
			stat_exchange.cas_data[8 + j/2] = (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		for (j=1; j<16; j=j+2) {
			tmp = (pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] >> 8) & 0xf;
			stat_exchange.cas_data[8 + j/2] |= (unsigned char) tmp;
			//	printk("j=%d, data=0x%x, i=%d, tmp = 0x%x\n", j, pq->bar1[(0x40 * stat_exchange.chan_num) + CAS_BIT_BUFFER_OFFSET + j] , j/2, tmp);
		}

		if(copy_to_user((void *) arg, (void *) &stat_exchange, sizeof(stat_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_RESET:
		module_printk(KERN_WARNING, "PTMC_E1_RESET.\n");
		module_printk(KERN_WARNING, "Unimplemented/n");
		return -ENOSYS;
		break;
	case PTMC_E1_BOARD_CONFIGURE:
		module_printk(KERN_WARNING, "PTMC_E1_BOARD_CONFIGURE.\n");
		module_printk(KERN_WARNING, "Unimplemented/n");
		return -ENOSYS;
		break;
	case PTMC_E1_GET_PCI_SLOT:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_SLOT.\n");
		ulTmp = PCI_SLOT( (unsigned int) pq->pci_device->devfn);
		if (copy_to_user((void *) arg, (void *) &ulTmp, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_PCI_BUSNUM:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_BUSNUM.\n");
		ulTmp = pq->pci_device->bus->number;
		if (copy_to_user((void *) arg, (void *) &ulTmp, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_READ_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR0.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pass_data.rdata = pq->bar0[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data)) )
		{
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR0.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pq->bar0[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_READ_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR1.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pass_data.rdata = pq->bar1[pass_data.raddress];
		if (copy_to_user((void *) arg, (void *) &pass_data, sizeof(pass_data)) )
		{
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR1.\n");
		if (copy_from_user((void *) &pass_data, (void *) arg, sizeof(pass_data)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF) return -EINVAL;
		pq->bar1[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_CAS_XMIT_PASS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CAS_XMIT_PASS\n");
		if(copy_from_user((void *) &cas_data, (void *) arg, _IOC_SIZE(cmd)))
		{
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}

		if ((cas_data.chan_num & (~DIRECT_IN)) > pq->e1_amount) return -EINVAL;

		if (copy_to_user((void *) arg, (void *) &cas_data, sizeof(cas_data)) )
		{
			return -EFAULT;
		}

		for (j=2; j<16; j=j+2) { //j==2 -> abcd2 -> cdata[1]
			tmp = pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xff0f;
			tmp |= cas_data.cdata[j/2] & 0xf0;
			pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}

		for (j=1; j<16; j=j+2) {
			tmp = pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xff0f;
			tmp |= ((cas_data.cdata[j/2] & 0x0f) << 4);
			pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}
		for (j=2; j<16; j=j+2) {
			tmp = pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xfff0;
			tmp |= ((cas_data.cdata[8 + j/2] & 0xf0) >> 4);
			pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;

		}
		for (j=1; j<16; j=j+2) {
			tmp = pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j];
			tmp &= 0xfff0;
			tmp |= (cas_data.cdata[8 + j/2] & 0x0f);
			pq->bar1[(0x40 * cas_data.chan_num) + CAS_BIT_BUFFER_OFFSET + j] = tmp;
		}

		break;
	case PTMC_E1_GET_VERSION:
		module_dprintk(KERN_DEBUG, "PTMC_E1_VERSION.\n");
		strncpy (version_exchange.proto_version, ecfg_proto_version, sizeof (version_exchange.proto_version));
		version_exchange.hw_fw_rev = pq->bar0[BAR0_HW_FW_REV];
		strncpy (version_exchange.driver_version, QUASAR_VERSION, sizeof (version_exchange.driver_version));
		version_exchange.board_id = pq->pci_device->device;
		if(copy_to_user((void *) arg, (void *) &version_exchange, sizeof(version_exchange)))
		{
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;



		/*
		 case PTMC_E1_PASS_ABCD_ZAP_FAKE:
		 //printk("PTMC_E1_PASS_ABCD_ZAP_FAKE\n");
		 if(copy_from_user((void *) &tmp, (void *) arg, _IOC_SIZE(cmd)))
		 {
		 module_printk(KERN_ERR, "Memory access fault!\n");
		 return -EFAULT;
		 }
		 pq->test_sig = tmp;
		 //printk("Passing:0x%x\n", tmp);
		 break;
		 */
	default:
		return -ENOTTY;
		break;
	}
	return 0;
}

/*
 static void check_bit(..)

 struct quasar_pmc * pmc - pointer to pmc device
 unsigned char chan_num - channel number
 unsigned char page - framer page 
 unsigned char line - framer register
 unsigned char smask - bit position in configuration uchar
 unsigned char hmask - bit position in hardware register
 unsigned char conf - data to write
 unsigned char *from_hw - readback
*/

static void check_bit_quasar8(struct quasar *pq,
		      unsigned char chan_num, unsigned char page,
		      unsigned char line, unsigned char smask,
		      const unsigned char hmask, unsigned char conf,
		      unsigned char *from_hw)
{
	unsigned int tmp;
	unsigned char sstore, htore, save_smask;

	tmp = framer_read(pq, chan_num, page, line);
	htore = (unsigned char) (tmp & hmask);	/* prepare for comparison curent value from hw */
	save_smask = smask;
	sstore = conf & smask;	/* prepare for comparison value from config */
	if (smask > hmask) {	/* allign software value to hardware position */
		while (smask > hmask) {
			smask >>= 1;
			sstore >>= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore <<= 1;
		}
	}
	if (htore ^ sstore) {	/* here sstore shifted to hardware position content of configuratuon value */
		tmp = framer_read(pq, chan_num, page, line);
		tmp &= ~hmask;
		tmp |= (sstore & hmask);
		framer_write(pq, chan_num, page, line, tmp);
	}
	/* reading back from hardware */
	tmp = framer_read(pq, chan_num, page, line);
	sstore = (unsigned char) (tmp & hmask);
	smask = save_smask;
	if (smask > hmask) {	/* allign hardware to software position */
		while (smask > hmask) {
			smask >>= 1;
			sstore <<= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore >>= 1;
		}
	}
	if (from_hw) {
		*from_hw |= sstore;
	}
}

/* Only allign to bit position in interface structure */
static void align_to_swposition_quasar8(const unsigned int tmp,
				unsigned char smask,
				const unsigned char hmask,
				unsigned char *aligned)
{
	unsigned char sstore;
	sstore = (unsigned char) (tmp & hmask);
	if (smask > hmask) {	/* allign hardware to software position */
		while (smask > hmask) {
			smask >>= 1;
			sstore <<= 1;
		}
	} else if (smask < hmask) {
		while (smask < hmask) {
			smask <<= 1;
			sstore >>= 1;
		}
	}
	if (aligned) {
		*aligned |= sstore;
	}
}

static void framer2ecfg(struct quasar *pq, unsigned char channo,
			unsigned long framer_reg, unsigned char ecfg_mask,
			unsigned char wcfg, unsigned char *rcfg)
{
	unsigned int tmp;

	if (channo & DIRECT_IN) {
		channo &= (~DIRECT_IN);
		tmp = framer_read(pq, channo, PMC_GPAGE(framer_reg),PMC_GREG(framer_reg));
		align_to_swposition_quasar8(tmp, ecfg_mask, PMC_GBIT(framer_reg), rcfg);
	} else {
		check_bit_quasar8(pq, channo, PMC_GPAGE(framer_reg),
			  PMC_GREG(framer_reg), ecfg_mask,
			  PMC_GBIT(framer_reg), wcfg, rcfg);
	}
}

/*
 Return codes:
 ENOTTY - invalid ioctl;
 EINVAL - passing invalid parameters;
 EFAULT - error copying data from userspace;
 0 - success;
*/
static int ioctl_quasar8(struct inode *inode_p, struct file *file_p,
			unsigned int cmd, unsigned long arg)
{
	int k, j;
	unsigned int tmp = 0, channo = 0;
	unsigned long ulTmp=0;
	struct pmc_config_chan wcfg, rcfg;
	struct pmc_err_ins err_ins;
	struct pmc_data_pass pass_data;
	struct pmc_status_chan stat;
	struct pmc_config_board bcfg;
	struct pmc_cas_xmit_data cas;
	struct pmc_version version;
	struct quasar *pq =
	    (struct quasar *) file_p->private_data;

	if (_IOC_TYPE(cmd) != PTMC_E1_8_MAGIC)
		return -ENOTTY;
	if (_IOC_NR(cmd) > PTMC_E1_8_MCN)
		return -ENOTTY;

	memset((void *) &bcfg, 0, sizeof(struct pmc_config_board));
	memset((void *) &rcfg, 0, sizeof(struct pmc_config_chan));
	memset((void *) &wcfg, 0, sizeof(struct pmc_config_chan));
	memset((void *) &err_ins, 0, sizeof(struct pmc_err_ins));
	memset((void *) &stat, 0, sizeof(struct pmc_status_chan));
	memset((void *) &cas, 0, sizeof(struct pmc_cas_xmit_data));

	switch (cmd) {
	case PTMC_E1_CHAN_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CHAN_CONFIGURE.\n");
		if (copy_from_user
		    ((void *) &wcfg, (void *) arg, sizeof(wcfg))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = wcfg.chan_num;
		if ((channo & (~DIRECT_IN)) > pq->e1_amount)
			return -EINVAL;

		/* CFG0 */
		framer2ecfg(pq, channo, CFG0_CAS_HP, CFG0_CAS, wcfg.cfg0,
			    &rcfg.cfg0);
		framer2ecfg(pq, channo, CFG0_RCV_TR_HP, CFG0_RCV_TR,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pq, channo, CFG0_XMIT_TR_HP, CFG0_XMIT_TR,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pq, channo, CFG0_RCV_HDB3_HP, CFG0_RCV_HDB3,
			    wcfg.cfg0, &rcfg.cfg0);
		framer2ecfg(pq, channo, CFG0_XMIT_HDB3_HP, CFG0_XMIT_HDB3,
			    wcfg.cfg0, &rcfg.cfg0);
		/* CFG1 */
		framer2ecfg(pq, channo, CFG1_NAT0_HP, CFG1_NAT0,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_NAT1_HP, CFG1_NAT1,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_NAT2_HP, CFG1_NAT2,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_NAT3_HP, CFG1_NAT3,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_NAT4_HP, CFG1_NAT4,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_INAT0_HP, CFG1_INAT0,
			    wcfg.cfg1, &rcfg.cfg1);
		framer2ecfg(pq, channo, CFG1_INAT1_HP, CFG1_INAT1,
			    wcfg.cfg1, &rcfg.cfg1);
		/* CFG2 */
		framer2ecfg(pq, channo, CFG2_RAI_AUTO_ON_HP,
			    CFG2_RAI_MANUAL, wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_RAI_ENA_HP, CFG2_RAI_ON,
			    wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_RCAS_AUTO_ON_HP,
			    CFG2_RCAS_MANUAL, wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_RCAS_ENA_HP, CFG2_RCAS_ON,
			    wcfg.cfg2, &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_X1_HP, CFG2_X1, wcfg.cfg2,
			    &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_X2_HP, CFG2_X2, wcfg.cfg2,
			    &rcfg.cfg2);
		framer2ecfg(pq, channo, CFG2_X3_HP, CFG2_X3, wcfg.cfg2,
			    &rcfg.cfg2);
		/* CFG3 */
		framer2ecfg(pq, channo, CFG3_CRC4_TR_HP, CFG3_CRC4_TR,
			    wcfg.cfg3, &rcfg.cfg3);
		framer2ecfg(pq, channo, CFG3_CRC4_DIS_HP, CFG3_CRC4_DIS,
			    wcfg.cfg3, &rcfg.cfg3);
		framer2ecfg(pq, channo, CFG3_CRC4_MANUAL_HP,
			    CFG3_CRC4_MANUAL, wcfg.cfg3, &rcfg.cfg3);
		/* CFG4 */
		wcfg.cfg4 ^= CFG4_TX_DIS;
		framer2ecfg(pq, channo, CFG4_TX_DIS_HP, CFG4_TX_DIS, wcfg.cfg4,
			    &rcfg.cfg4);
		rcfg.cfg4 ^= CFG4_TX_DIS;
		framer2ecfg(pq, channo, CFG4_AIS_HP, CFG4_AIS, wcfg.cfg4,
			    &rcfg.cfg4);
		framer2ecfg(pq, channo, CFG4_LLOOP_HP, CFG4_LLOOP,
			    wcfg.cfg4, &rcfg.cfg4);
		framer2ecfg(pq, channo, CFG4_RLOOP_HP, CFG4_RLOOP,
			    wcfg.cfg4, &rcfg.cfg4);
		framer2ecfg(pq, channo, CFG4_PLOOP_HP, CFG4_PLOOP,
			    wcfg.cfg4, &rcfg.cfg4);

		/* Write port configuration */
		if (!(channo & DIRECT_IN)) {
			/* configure board part */
			if (wcfg.cfg0 & CFG0_MASTER) {
				pq->bar0[BAR0_BSLS] |= (unsigned int) (1 << channo);	/* restore bs/ls */
			} else {	/*allow to sync to this chan */
				pq->bar0[BAR0_BSLS] &= (unsigned int) ~(1 << channo);	/* restore bs/ls */
			}
		}
		/* Read port configuration */
		channo &= (~DIRECT_IN);

		/* configure board part */
		/* get BS/LS */
		tmp = (unsigned char) pq->bar0[BAR0_BSLS];
		align_to_swposition_quasar8(tmp, CFG0_MASTER, 1 << channo, &rcfg.cfg0);	/* 0-6-chan_dep */


		/* get TXAO */
		tmp = ~(unsigned char) (pq->bar0[BAR0_TXAO]);
		align_to_swposition_quasar8(tmp, CFG4_TX_DIS, 1 << channo, &rcfg.cfg4);	/* 0-4-chan_dep */

		if (copy_to_user
		    ((void *) arg, (void *) &rcfg, sizeof(rcfg))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		break;
	case PTMC_E1_ERR_EMU:
		module_dprintk(KERN_DEBUG, "PTMC_E1_ERR_EMU.\n");

		if (copy_from_user
		    ((void *) &err_ins, (void *) arg, sizeof(err_ins))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = err_ins.chan_num;
		if (channo > pq->e1_amount)
			return -EINVAL;

		/* generate errors */
		framer2ecfg(pq, channo, ERR_EMU0_LOF_NFAS_HP,
			    ERR_EMU0_LOF_NFAS, err_ins.pmc_err, NULL);
		framer2ecfg(pq, channo, ERR_EMU0_LOF_FAS_HP,
			    ERR_EMU0_LOF_FAS, err_ins.pmc_err, NULL);
		framer2ecfg(pq, channo, ERR_EMU0_CRC4_HP, ERR_EMU0_CRC4,
			    err_ins.pmc_err, NULL);
		/* self restore zero state; */
		framer2ecfg(pq, channo, ERR_EMU0_LOF_NFAS_HP,
			    ERR_EMU0_LOF_NFAS, 0, NULL);
		framer2ecfg(pq, channo, ERR_EMU0_LOF_FAS_HP,
			    ERR_EMU0_LOF_FAS, 0, NULL);
		framer2ecfg(pq, channo, ERR_EMU0_CRC4_HP, ERR_EMU0_CRC4,
			    0, NULL);

		break;
	case PTMC_E1_GET_STATS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_STATS.\n");

		if (copy_from_user
		    ((void *) &stat, (void *) arg, sizeof(stat))) {
			module_printk(KERN_ERR, "Memory access fault!");
			return -EFAULT;
		}
		channo = stat.chan_num & (~DIRECT_IN);
		if (channo > pq->e1_amount)
			return -EINVAL;

		framer2ecfg(pq, channo, STAT0_LOS_HP, STAT0_LOS, 0,
			    &stat.stat0);
		framer2ecfg(pq, channo, STAT0_LOF_HP, STAT0_LOF, 0,
			    &stat.stat0);
		framer2ecfg(pq, channo, STAT0_CAS_LOM_HP, STAT0_CAS_LOM,
			    0, &stat.stat0);
		framer2ecfg(pq, channo, STAT0_CRC4_SYNC_HP,
			    STAT0_CRC4_SYNC, 0, &stat.stat0);
		framer2ecfg(pq, channo, STAT0_AIS_HP, STAT0_AIS, 0,
			    &stat.stat0);
		framer2ecfg(pq, channo, STAT0_RAIS_HP, STAT0_RAIS, 0,
			    &stat.stat0);
		tmp = framer_read(pq, channo, 0x3, 0x10);
		if (!((tmp & PMC_GBIT(STAT0_E_BIT1_HP))
		     && (tmp & PMC_GBIT(STAT0_E_BIT2_HP))))
			stat.stat0 |= STAT0_E_BIT;

		tmp = framer_read(pq, channo, 0x3, 0x11);
		if ((tmp & PMC_GBIT(STAT0_CRC4_1_HP))
		    || (tmp & PMC_GBIT(STAT0_CRC4_2_HP)))
			stat.stat0 |= STAT0_CRC4;

		framer2ecfg(pq, channo, STAT1_NAT0_HP, STAT1_NAT0, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_NAT1_HP, STAT1_NAT1, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_NAT2_HP, STAT1_NAT2, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_NAT3_HP, STAT1_NAT3, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_NAT4_HP, STAT1_NAT4, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_INAT0_HP, STAT1_INAT0, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_INAT1_HP, STAT1_INAT1, 0,
			    &stat.stat1);
		framer2ecfg(pq, channo, STAT1_SLIP_HP, STAT1_SLIP, 0,
			    &stat.stat1);

		framer2ecfg(pq, channo, STAT2_Y_HP, STAT2_Y, 0,
			    &stat.stat2);
		framer2ecfg(pq, channo, STAT2_X1_HP, STAT2_X1, 0,
			    &stat.stat2);
		framer2ecfg(pq, channo, STAT2_X2_HP, STAT2_X2, 0,
			    &stat.stat2);
		framer2ecfg(pq, channo, STAT2_X3_HP, STAT2_X3, 0,
			    &stat.stat2);

		framer2ecfg(pq, channo, OVR_HDB3_HP, OVR_HDB3, 0,
			    &stat.ovr);
		framer2ecfg(pq, channo, OVR_FAS_HP, OVR_FAS, 0,
			    &stat.ovr);
		framer2ecfg(pq, channo, OVR_CRC4_HP, OVR_CRC4, 0,
			    &stat.ovr);

		/* reading FAS counter */
		stat.fas = framer_read_pr(pq, channo, CNT_FAS_HP);
		/* reading HDB3 counter */
		stat.hdb3 = framer_read_pr(pq, channo, CNT_BPVHI_HP);
		stat.hdb3 <<= 8;
		stat.hdb3 |= framer_read_pr(pq, channo, CNT_BPVLO_HP);
		/* reading CRC4 counter */
		stat.crc4 = framer_read_pr(pq, channo, CNT_CRC4HI_HP);
		stat.crc4 <<= 8;
		stat.crc4 |= framer_read_pr(pq, channo, CNT_CRC4LO_HP);

		/* clear counter if requested */
		if (stat.mgmt & OVR_FAS) {
			framer_write_pr(pq, channo, CNT_FAS_HP, 0);
		}
		if (stat.mgmt & OVR_HDB3) {
			framer_write_pr(pq, channo, CNT_BPVHI_HP, 0);
			framer_write_pr(pq, channo, CNT_BPVLO_HP, 0);
		}
		if (stat.mgmt & OVR_CRC4) {
			framer_write_pr(pq, channo, CNT_CRC4HI_HP, 0);
			framer_write_pr(pq, channo, CNT_CRC4LO_HP, 0);
		}

		/* get cas signalling */
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pq, channo, 0x9, 0x10 + j);
			if (j < (2 * k + 1)) {
				stat.cas_data[k] =
				    (unsigned char) (tmp & 0xf);
				stat.cas_data[k] <<= 4;

			} else {
				tmp &= 0x0f;
				stat.cas_data[k] |= (unsigned char) tmp;
				k += 1;
			}
		}
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pq, channo, 0xA, 0x10 + j);
			if (j < (2 * k + 1)) {
				stat.cas_data[8 + k] =
				    (unsigned char) (tmp & 0xf);
				stat.cas_data[8 + k] <<= 4;
			} else {
				tmp &= 0x0f;
				stat.cas_data[8 + k] |=
				    (unsigned char) tmp;
				k += 1;
			}
		}

		if (copy_to_user
		    ((void *) arg, (void *) &stat, sizeof(stat))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_RESET:
		/* Reset board. Not tested very well. */
		module_dprintk(KERN_DEBUG, "PTMC_E1_RESET.\n");

		/* initialising board registers */

		QUASAR_HAL(init_core_regs(pq));

		/* initialising framer registers */
		QUASAR_HAL(init_framer_regs(pq));

		/* initialising time-slots maping table */
		QUASAR_HAL(init_switch_map(pq));
		break;
	case PTMC_E1_BOARD_CONFIGURE:
		module_dprintk(KERN_DEBUG, "PTMC_E1_BOARD_CONFIGURE.\n");

		if (copy_from_user((void *) &bcfg, (void *) arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		/* reading ctrl register */
		tmp = pq->bar0[BAR0_CTRL];

		tmp &= ~(CTRL_NR1EN | CTRL_NR2EN | 
			CTRL_NR1USE | CTRL_NR2USE |
			CTRL_ABSEL | CTRL_MASL);
			
		if (bcfg.mode1 & MODE1_NF1_GEN_ENA) {
			tmp |= CTRL_NR1EN;
		}
		if (bcfg.mode1 & MODE1_NF2_GEN_ENA) {
			tmp |= CTRL_NR2EN;
		}
		if (bcfg.mode1 & MODE1_NF1_USE_ENA) {
			tmp |= CTRL_NR1USE;
		}
		if (bcfg.mode1 & MODE1_NF2_USE_ENA) {
			tmp |= CTRL_NR2USE;
		}
		if (bcfg.mode1 & MODE1_AB) {
			tmp |= CTRL_ABSEL;
		}
		if (bcfg.mode1 & MODE1_MA_SL_SEL) {
			tmp |= CTRL_MASL;
		}
		pq->bar0[BAR0_CTRL] = tmp;
		udelay(2000);
		break;
	case PTMC_E1_GET_PCI_SLOT:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_SLOT.\n");

		ulTmp = PCI_SLOT((unsigned int) pq->pci_device->devfn);
		if (copy_to_user((void *)arg, (void *)&ulTmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_PCI_BUSNUM:
		module_dprintk(KERN_DEBUG, "PTMC_E1_GET_PCI_BUSNUM.\n");

		ulTmp = pq->pci_device->bus->number;
		if (copy_to_user((void *)arg, (void *)&ulTmp, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
	case PTMC_E1_READ_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR0.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = pq->bar0[pass_data.raddress];
		if (copy_to_user((void *)arg, (void *)&pass_data, sizeof(pass_data))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR0:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR0.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pq->bar0[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_READ_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_READ_BAR1.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pass_data.rdata = pq->bar1[pass_data.raddress];
		if (copy_to_user((void *)arg, (void *)&pass_data, sizeof(pass_data))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_WRITE_BAR1:
		module_dprintk(KERN_DEBUG, "PTMC_E1_WRITE_BAR1.\n");

		if (copy_from_user((void *)&pass_data, (void *)arg, sizeof(pass_data))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		if (pass_data.raddress > 0xFFFF)
			return -EINVAL;
		pq->bar1[pass_data.raddress] = pass_data.rdata;
		break;
	case PTMC_E1_CAS_XMIT_PASS:
		module_dprintk(KERN_DEBUG, "PTMC_E1_CAS_XMIT_PASS\n");

		if (copy_from_user((void *)&cas, (void *)arg, _IOC_SIZE(cmd))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		channo = cas.chan_num;
		if ((channo & (~DIRECT_IN)) > pq->e1_amount)
			return -EINVAL;

		if (!(channo & DIRECT_IN)) {	/* writing cas data */
			channo &= (~DIRECT_IN);
			module_dprintk(KERN_DEBUG,
				       "Pass data! Chan_num = %d\n",
				       channo);

			k = 0;
			for (j = 0; j < 16; j++) {
				if (j < (2 * k + 1)) {
					tmp = (cas.cdata[k] >> 4) & 0x0f;
					framer_write(pq, channo, 0x5,
						     0x10 + j, tmp);
				} else {
					tmp = (cas.cdata[k] & 0x0f);
					framer_write(pq, channo, 0x5,
						     0x10 + j, tmp);
					k += 1;
				}
			}
			k = 0;
			for (j = 0; j < 16; j++) {
				if (j < (2 * k + 1)) {
					tmp =
					    (cas.cdata[8 + k] >> 4) & 0x0f;
					framer_write(pq, channo, 0x6,
						     0x10 + j, tmp);
				} else {
					tmp = (cas.cdata[8 + k] & 0x0f);
					framer_write(pq, channo, 0x6,
						     0x10 + j, tmp);
					k += 1;
				}
			}
			memset((void *) &cas, 0, sizeof(struct pmc_cas_xmit_data));	/* clear to get values back */
		}
		/* Reading back cas signalling */
		channo &= (~DIRECT_IN);
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pq, channo, 0x5, 0x10 + j);
			if (j < (2 * k + 1)) {
				tmp &= 0x0f;
				cas.cdata[k] = (unsigned char) tmp;
				cas.cdata[k] <<= 4;
			} else {
				tmp &= 0x0f;
				cas.cdata[k] |= (unsigned char) tmp;
				k += 1;
			}

		}
		k = 0;
		for (j = 0; j < 16; j++) {
			tmp = framer_read(pq, channo, 0x6, 0x10 + j);
			if (j < (2 * k + 1)) {
				tmp &= 0x0f;
				cas.cdata[8 + k] = (unsigned char) tmp;
				cas.cdata[8 + k] <<= 4;
			} else {
				tmp &= 0x0f;
				cas.cdata[8 + k] |= (unsigned char) tmp;
				k += 1;
			}
		}

		if (copy_to_user((void *) arg, (void *) &cas, sizeof(cas))) {
			return -EFAULT;
		}
		break;
	case PTMC_E1_GET_VERSION:
		module_dprintk(KERN_DEBUG, "PTMC_E1_VERSION.\n");
		strncpy(version.proto_version, ecfg_proto_version,
			sizeof(version.proto_version));
		version.hw_fw_rev = pq->bar0[BAR0_HW_FW_REV];
		version.board_id = pq->pci_device->device;
		strncpy(version.driver_version, QUASAR_VERSION,
			sizeof(version.driver_version));
		if (copy_to_user
		    ((void *) arg, (void *) &version, sizeof(version))) {
			module_printk(KERN_ERR, "Memory access fault!\n");
			return -EFAULT;
		}
		break;
		/*
		   case PTMC_E1_PASS_ABCD_ZAP_FAKE: removed.
		 */
	default:
		break;
	}
	return 0;
}

static void quasar_find_master (struct quasar_dahdi * pd)
{
	int i;
	struct quasar_span *ts;
	struct quasar * pq = pd->pvt;
	int syncsrc, master;

	master = pq->bar0[BAR0_CTRL] & 0x0f;
	syncsrc = pd->spans[master]->span.spanno;
	ts = pd->spans[master];

	if (pq->bar0[BAR0_CTRL] & (CTRL_HLD | CTRL_FREE))
		syncsrc = 0; //Using internal sync source for CT-BUS

	for (i=0; i<pd->num_spans; ++i) {
		ts = pd->spans[i];
		ts->span.syncsrc = syncsrc;
	}
}

static void check_alarm_quasarm (struct quasar_dahdi * pd)
{
	int i;
	struct quasar_span *qspan;
	struct quasar * pq = pd->pvt;

	for (i=0; i<pd->num_spans; ++i) {
		qspan = pd->spans[i];

		if (pmc_reg_check(pq, i, FRS_ADDR_OFFSET, FRS_RAI_RX)) {
			qspan->span.alarms |= DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		}
		if (pmc_reg_check(pq, i, FRS_ADDR_OFFSET, FRS_RED)) { //LOF
			qspan->span.alarms |= DAHDI_ALARM_RED;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_RED;
		}
		if (pmc_reg_check(pq, i, FRS_ADDR_OFFSET, FRS_AIS)) {
			qspan->span.alarms |= DAHDI_ALARM_BLUE;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_BLUE;
		}
		if (qspan->unframed) {
			qspan->span.alarms = 0;
		}
		dahdi_alarm_notify (&qspan->span);
        DAHDI_COUNT(qspan->span, fascount, count.fas) = (unsigned short)pmc_reg_read (pq, i, FRS_FRAMEERR_OFFSET);
		DAHDI_COUNT(qspan->span, bpvcount, count.bpv) = (unsigned short)pmc_reg_read (pq, i, FRS_HDB3ERR_OFFSET);
		DAHDI_COUNT(qspan->span, crc4count, count.crc4) = (unsigned short)pmc_reg_read (pq, i, FRS_CRC4ERR_OFFSET);
#ifdef HAVE_DAHDI_DEVICE
                pq->ddev->irqmisses = pd->irqmisses;
#else
		qspan->span.irqmisses = pd->irqmisses;
#endif //HAVE_DAHDI_DEVICE
        }
}

static void check_alarm_quasar8(struct quasar_dahdi *pd)
{
	int i;
    unsigned long reg;
	struct quasar_span *qspan;
	struct quasar *pq = pd->pvt;

	for (i=0; i < pd->num_spans; ++i) {
		qspan = pd->spans[i];

		if (framer_checkbit(pq, i, STAT0_RAIS_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		}
		if (framer_checkbit(pq, i, STAT0_LOF_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_RED;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_RED;
		}
		if (framer_checkbit(pq, i, STAT0_AIS_HP)) {
			qspan->span.alarms |= DAHDI_ALARM_BLUE;
			qspan->span.alarms &= ~DAHDI_ALARM_YELLOW;
		} else {
			qspan->span.alarms &= ~DAHDI_ALARM_BLUE;
		}
		dahdi_alarm_notify(&qspan->span);

		DAHDI_COUNT(qspan->span, fascount, count.fas) =
			framer_read_pr(pq, i, CNT_FAS_HP);

		reg = framer_read_pr(pq, i, CNT_BPVHI_HP);
		reg <<= 8;
		reg |= framer_read_pr(pq, i, CNT_BPVLO_HP);
		DAHDI_COUNT(qspan->span, bpvcount, count.bpv) = reg;

		reg = framer_read_pr(pq, i, CNT_CRC4HI_HP);
		reg <<= 8;
		reg |= framer_read_pr(pq, i, CNT_CRC4LO_HP);
		DAHDI_COUNT(qspan->span, crc4count, count.crc4) = reg;

#ifdef HAVE_DAHDI_DEVICE
		pq->ddev->irqmisses = pd->irqmisses;
#else
		qspan->span.irqmisses = pd->irqmisses;
#endif //HAVE_DAHDI_DEVICE

		if (framer_checkbit(pq, i, STAT0_LOF_HP)) {
			qspan->frok = 0; /* Frame is bad */
		} else {
			qspan->frok++;
		}
		if (!qspan->frok || 8 == qspan->frok) {
			/* Reset equalizer if LOF detected and 1s after LOF finished. */
			reg = framer_read(pq, i, 0x1, 0x1F);
			framer_write(pq, i, 0x1, 0x1F, reg&(~(0x18))); /* switch of auto equalizer */
			framer_write(pq, i, 0x1, 0x1F, reg|0x18); /* switch on auto equalizer */
		}
	}
}
static void clearstat_quasarm (struct quasar_dahdi * pd) {
	struct quasar * pq = pd->pvt;
	struct quasar_span *qspan;
	int i;

	for (i=0; i<pd->num_spans; ++i) {
		qspan = pd->spans[i];

		if (0 == qspan->clear_counters) continue; //Already cleared
		--qspan->clear_counters;
		if (0 != qspan->clear_counters) continue;
		module_dprintk(KERN_DEBUG, "e12zap_clearstat:"
			       "Clearing span %d error counters\n", qspan->span.spanno);
		
		pmc_reg_write (pq, i, FRS_FRAMEERR_OFFSET, 0);
		pmc_reg_write (pq, i, FRS_HDB3ERR_OFFSET, 0);
		pmc_reg_write (pq, i, FRS_CRC4ERR_OFFSET, 0);

		pd->irqmisses = 0;
	}
}

static void clearstat_quasar8 (struct quasar_dahdi * pd) {
	struct quasar * pq = pd->pvt;
	struct quasar_span *qspan;
	int i;

	for (i=0; i<pd->num_spans; ++i) {
		qspan = pd->spans[i];

		if (0 == qspan->clear_counters) continue; //Already cleared
		--qspan->clear_counters;
		if (0 != qspan->clear_counters) continue;
		module_dprintk(KERN_DEBUG, "e12zap_clearstat:"
			       "Clearing span %d error counters\n", qspan->span.spanno);
		
		framer_write_pr(pq, i, CNT_FAS_HP, 0);
		framer_write_pr(pq, i, CNT_BPVHI_HP, 0);
		framer_write_pr(pq, i, CNT_BPVLO_HP, 0);
		framer_write_pr(pq, i, CNT_CRC4HI_HP, 0);
		framer_write_pr(pq, i, CNT_CRC4LO_HP, 0);

		pd->irqmisses = 0;
	}
}

static struct quasar_span *qspan_from_dspan (struct dahdi_span *span) {
        return container_of(span, struct quasar_span, span);
}
static int maint_quasar8(struct dahdi_span *span, int cmd)
{
	module_printk(KERN_WARNING, "e21zap_maint: Use ecfg testing options.\n");
	return -ENOSYS;
}
static int maint_quasarm(struct dahdi_span *span, int cmd)
{
	int res = 0;
	unsigned char ucTmp;
	struct quasar_span * espan = qspan_from_dspan(span);
	struct quasar_dahdi * epmc = espan->pvt;
	struct quasar * pq = epmc->pvt;
	int channo = espan->spannumber;

	if (!pq->have_liu[channo]) {
		module_printk(KERN_WARNING, "e21zap_maint: Not supported on non-LUI port, yet.\n");
		return -ENOSYS;
	}
	acc_lxt_reg(pq, channo, LR_CR2, LIU_READ, &ucTmp);

	/* We don't support maint commands.
         Use emcfg instead, it's much more functional. */
	switch(cmd) {
	case DAHDI_MAINT_NONE:
                ucTmp = LR_CR2_ENLOOP; /* TODO */
		module_printk(KERN_WARNING, "e21zap_maint: Clearing all maint modes.\n");
		acc_lxt_reg(pq, channo, LR_CR2, LIU_WRITE, &ucTmp);
		break;
	case DAHDI_MAINT_LOOPUP:
		module_printk(KERN_WARNING, "e21zap_maint: Transmitting loopup code .\n"); //@ %d.%d
		ucTmp |= LR_CR2_EPAT0;
		ucTmp &= ~LR_CR2_EPAT1;
		acc_lxt_reg(pq, channo, LR_CR2, LIU_WRITE, &ucTmp);
		break;
	case DAHDI_MAINT_LOOPDOWN:
		module_printk(KERN_WARNING, "e21zap_maint: Transmitting loopdown code .\n"); //@ %d.%d
		ucTmp |= LR_CR2_EPAT0 |LR_CR2_EPAT1;
		acc_lxt_reg(pq, channo, LR_CR2, LIU_WRITE, &ucTmp);
		break;

	case DAHDI_MAINT_LOCALLOOP:
		/* loop=on, rloop=? */
	case DAHDI_MAINT_REMOTELOOP:
		/* loop=?, rloop=on */
		module_printk(KERN_WARNING, "e21zap_maint: Use ecfg testing options.\n");
		res = -ENOSYS;
		break;
	default:
		module_printk(KERN_ERR, "e21zap_maint: Unknown maint command: %d\n", cmd);
		res = -ENOTTY;
		break;
	}
	return res;
}

static int rbsbits_quasarm(struct dahdi_chan *chan, int bits)
{
	//trunk for CAS nibles...
	//put CAS nibbles to appropriate position
	struct quasar_span * pspan = chan->pvt;
	struct quasar_dahdi * pd = pspan->pvt;
	struct quasar * pq = pd->pvt;
	unsigned int tmp;

	module_dprintk(KERN_DEBUG, "e12zap_rbsbits:->%x on %s\n", bits, chan->name);
	if (chan->chanpos == 16) return 0;

    if(chan->chanpos < 16){
	    	//put data
		tmp = pq->bar1[pspan->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos];
		tmp &= 0xff0f;
		tmp |=  (bits << 4);
		pq->bar1[pspan->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos] = tmp;
	}
	else {
		//put data
		tmp = pq->bar1[pspan->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos - 16];
		tmp &= 0xfff0;
		tmp |= bits;
		pq->bar1[pspan->spannumber * 0x40 + CAS_BIT_BUFFER_OFFSET + chan->chanpos - 16] = tmp;
    }
	return 0;
}
static int rbsbits_quasar8(struct dahdi_chan *chan, int bits)
{
	//trunk for CAS nibles...
	//put CAS nibbles to appropriate position
	struct quasar_span * pspan = chan->pvt;
	struct quasar_dahdi * pd = pspan->pvt;
	struct quasar * pq = pd->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_rbsbits:->%x on %s\n", bits, chan->name);
	if (chan->chanpos == 16) return 0;

	if(chan->chanpos < 16){
		framer_write(pq, pspan->spannumber, 0x5, 0x10 + chan->chanpos, bits);
	}
	else {
		framer_write(pq, pspan->spannumber, 0x6, 0x10 + chan->chanpos - 16, bits);
	}
	return 0;
}

static int shutdown_quasar(struct dahdi_span *span)
{
	int wasrunning, i, tmp;
	unsigned char working=0;
	struct quasar_span *qspan = qspan_from_dspan(span);
	struct quasar_dahdi * pd = qspan->pvt;
	struct quasar * pq = pd->pvt;
	wasrunning = span->flags & DAHDI_FLAG_RUNNING;
	span->flags &= ~DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_shutdown\n");
	if (wasrunning){
		for(i=0;i<pd->num_spans;i++){
			if(pd->spans[i]->span.flags & DAHDI_FLAG_RUNNING)
				working =1;
		}
	}
	if (!working){
		//disable DMA on recv and transmit
		tmp = pq->bar0[BAR0_DMACR];
		tmp &= ~(DMA_WENA|DMA_RENA);
		pq->bar0[BAR0_DMACR] = tmp;

		//cancel pending interrupts
		pq->bar0[BAR0_IFR] = 0;

		//disable DMA interrupt, mask
		tmp = pq->bar0[BAR0_IMR];
		tmp &= ~(0x20000000); //DMA
		pq->bar0[BAR0_IMR] = tmp;
	}
	return 0;
}

static int startup_quasarm(struct dahdi_span *span)
{
	unsigned int i, tmp, ucTmp1;
	unsigned char ucTmp=0;
	struct quasar_span *qspan = qspan_from_dspan(span);
    struct quasar_dahdi *pd = qspan->pvt;
	struct quasar * pq = pd->pvt;
    int alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_startup: ");

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++)
	{
		memset(qspan->ec_chunk1[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(qspan->ec_chunk2[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	dprintk("", "Using ");

	if (qspan->unframed) {
	    if (pq->have_liu[qspan->spannumber]) {
    	    pq->bar1[(0x40*qspan->spannumber)+FRC1_ADDR_OFFSET] = 0;
    	} else {
			pq->bar1[(0x40*qspan->spannumber)+FRC1_ADDR_OFFSET] = FRC_HDB3_RX_EN | FRC_HDB3_TX_EN | 
				FRC_SHORT_HAUL_EN | FRC_TXEN;
		}
    }
	//set page 1;
        tmp = pq->bar1[qspan->spannumber * 0x40];

    if(span->lineconfig & DAHDI_CONFIG_HDB3) //set HDB3
	{
		if (tmp & FRC_SHORT_HAUL_EN) {
			pq->bar1[qspan->spannumber * 0x40] |= (FRC_HDB3_RX_EN | FRC_HDB3_TX_EN) ;
		} else {
			ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_ENCENB | LR_CR1_EC1 | LR_CR1_EC4; //HDB3, long haul, JA-tx;
			acc_lxt_reg(pq, qspan->spannumber, LR_CR1, LIU_WRITE, &ucTmp);
            ucTmp = 0;
            acc_lxt_reg(pq, qspan->spannumber, LR_CR1, LIU_READ, &ucTmp);
            if (!(ucTmp & LR_CR1_ENCENB)) 
                    		 module_printk(KERN_ERR, "LIU configuration error\n");
		} 
		dprintk("", "HDB3/");
 	}
    else //set AMI
	{
        if (tmp & FRC_SHORT_HAUL_EN) {
			pq->bar1[qspan->spannumber * 0x40] &= ~(FRC_HDB3_RX_EN | FRC_HDB3_TX_EN) ;
	 	} else {
			ucTmp = LR_CR1_JASEL0 | LR_CR1_UNIEB | LR_CR1_EC1 | LR_CR1_EC4; //AMI, long haul, JA-tx;
			acc_lxt_reg(pq, qspan->spannumber, LR_CR1, LIU_WRITE, &ucTmp);
            ucTmp = 0;
            acc_lxt_reg(pq, qspan->spannumber, LR_CR1, LIU_READ, &ucTmp);
            if (ucTmp & LR_CR1_ENCENB)
          		module_printk(KERN_ERR, "LIU configuration error\n");
		} 
		dprintk("", "AMI/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CCS)
	{       //set CSS
		pq->bar1[qspan->spannumber * 0x40] &= ~FRC_MFR_TX_EN;
		dprintk("", "CCS/");
	}
	else  //set CAS
	{
		pq->bar1[qspan->spannumber * 0x40] |= FRC_MFR_TX_EN;
		dprintk("", "CAS/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CRC4)
	{
		pq->bar1[qspan->spannumber * 0x40] |= FRC_CRC4_TX_EN;
		dprintk("", "CRC4 ");
	}
	else
	{
		pq->bar1[qspan->spannumber * 0x40] &= ~FRC_CRC4_TX_EN;
		dprintk("", "non-CRC4 ");
	}

        dprintk("", "on span %s, ", span->name);
	if (qspan->syncprior) //if !0 allow to use for syncronisation.
        {
	    tmp = pq->bar0[BAR0_BSLS]; //get bs/ls
	    ucTmp1 = ~(1<<qspan->spannumber);
	    tmp&= ucTmp1;
	    pq->bar0[BAR0_BSLS] = tmp; //restore bs/ls
            dprintk("", "can be the sync. source\n");
	}
	else
	{
	    tmp = pq->bar0[BAR0_BSLS]; //get bs/ls
	    ucTmp1 = (1<<qspan->spannumber);
	    tmp|= ucTmp1;
	    pq->bar0[BAR0_BSLS] = tmp; //restore bs/ls
            dprintk("", "can not be the sync. source\n");
	}

	if (!alreadyrunning) {
	    module_dprintk(KERN_DEBUG, "Enable DMA...\n");
	// Only if we're not already going
	//allow DMA on recv and transmit
	    tmp = pq->bar0[BAR0_DMACR];
	    if (QUASAR16(pq->pci_device)) tmp |= DMA_WENA;
		else tmp |= (DMA_WENA | DMA_RENA);
	    pq->bar0[BAR0_DMACR] = tmp;
	    //allow DMA interrupt
	    tmp = pq->bar0[BAR0_IMR];
	    tmp |= 0x20000000; //DMA
	    pq->bar0[BAR0_IMR] = tmp;
	    span->flags |= DAHDI_FLAG_RUNNING;
	}
	qspan->clear_counters = 3;
	return 0;
}
static int startup_quasar8(struct dahdi_span *span)
{
	int i;
	struct quasar_span *qspan = qspan_from_dspan(span);
	struct quasar_dahdi *pd = qspan->pvt;
	struct quasar *pmc = pd->pvt;
	int alreadyrunning = span->flags & DAHDI_FLAG_RUNNING;

	module_dprintk(KERN_DEBUG, "e12zap_startup: ");

	/* initialize the start value for the entire chunk of last ec buffer */
	for(i = 0; i < span->channels; i++) {
		memset(qspan->ec_chunk1[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
		memset(qspan->ec_chunk2[i],
			DAHDI_LIN2X(0,span->chans[i]),DAHDI_CHUNKSIZE);
	}
	dprintk("", "Using ");

	if(span->lineconfig & DAHDI_CONFIG_HDB3) {
		framer_setbit(pmc, qspan->spannumber, CFG0_RCV_HDB3_HP);
		framer_setbit(pmc, qspan->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "HDB3/");
	} else {
		framer_clrbit(pmc, qspan->spannumber, CFG0_RCV_HDB3_HP);
                framer_clrbit(pmc, qspan->spannumber, CFG0_XMIT_HDB3_HP);
		dprintk("", "AMI/");
	}

	if(span->lineconfig & DAHDI_CONFIG_CCS) {
		framer_setbit(pmc, qspan->spannumber, CFG0_CAS_HP);
		dprintk("", "CCS/");
	} else {
		framer_clrbit(pmc, qspan->spannumber, CFG0_CAS_HP);
		dprintk("", "CAS/");
	}

	/* Basic CRC-4 synchronization processing. */
	framer_clrbit(pmc, qspan->spannumber, CFG3_CRC4_DIS_HP);
	/* Automatic CRC-interworking */
	framer_clrbit(pmc, qspan->spannumber, CFG3_CRC4_MANUAL_HP);

	/* Should I strict DAHDI's readme. See quasar/crc4.txt notes. */
	if (span->lineconfig & DAHDI_CONFIG_CRC4) {
		dprintk("", "CRC4 ");
	} else {
		dprintk("", "non-CRC4 ");
	}

	dprintk("", "on span %s, ", span->name);

	i = pmc->bar0[BAR0_BSLS]; /* get bs/ls */
	if (qspan->syncprior) {
		i &= ~(1<<qspan->spannumber);
		dprintk("", "can be the sync. source\n");
	} else {
		i |= (1<<qspan->spannumber);
		dprintk("", "can not be the sync. source\n");
	}
	pmc->bar0[BAR0_BSLS] = i; /* restore bs/ls */

	if (!alreadyrunning) {
		module_dprintk(KERN_DEBUG, "Enable DMA...\n");
                /* Enable receive/transmit DMA if not enabled. */
		i = pmc->bar0[BAR0_DMACR];
		i |= (DMA_WENA | DMA_RENA);
		pmc->bar0[BAR0_DMACR] = i;
		/* allow DMA interrupts, unmask, allow CASsig interrupt */
		i = pmc->bar0[BAR0_IMR];
		i |= 0x20000000;
		pmc->bar0[BAR0_IMR] = i;
		span->flags |= DAHDI_FLAG_RUNNING;
	}
	qspan->clear_counters = 3;

	return 0;
}
static void stop_dma(struct quasar *pq) 
{

	unsigned int tmp;

	//disable DMA on recv and transmit
	tmp = pq->bar0[BAR0_DMACR];
	tmp &= ~(DMA_WENA|DMA_RENA);
	pq->bar0[BAR0_DMACR] = tmp;

	//disable DMA interrupt, mask
	tmp = pq->bar0[BAR0_IMR];
	tmp &= ~(0x20000000);
	pq->bar0[BAR0_IMR] = tmp;
}

static void rbs_delivery_quasarm(struct quasar * pq)
{
	struct quasar_dahdi *pd = pq->priv_data;
	unsigned int spanid, chanid, tmp1;
	unsigned int rxcas[0x40];	//pointers to map CAS signalling. An ugly hack. TODO: rework.

	for (spanid=0; spanid < pd->num_spans; spanid++) {
		volatile unsigned int *casbar = &(pq->bar1[spanid * 0x40 + CAS_BIT_BUFFER_OFFSET]);

		if(!(pd->spans[spanid]->span.flags & DAHDI_FLAG_RUNNING))
			continue;
		if (pd->spans[spanid]->span.lineconfig & DAHDI_CONFIG_CCS)
			continue;
		if (spanid%4 != pd->intcount%4) /* RBS decimation. 0,4 1,5 2,6 3,7 */
			continue;

		memcpy(rxcas, (void *)casbar, 0x40);

		for(chanid=1; chanid < 16; chanid++) //chans < 16                            dahdi channels [1:15]  - chans [0:14]
		{
			tmp1 = (rxcas[chanid] >> 12 ) & 0x0f;
			if (tmp1 != pd->spans[spanid]->span.chans[chanid-1]->rxsig)
			{
				if(!(pd->spans[spanid]->span.chans[chanid-1]->sig & DAHDI_SIG_CLEAR))
				{
					dahdi_rbsbits(pd->spans[spanid]->span.chans[chanid-1], tmp1);
				}
			}
		}
		
		for(chanid=1;chanid<16;chanid++) //chans > 16                            dahdi channels [17:31] -chans [16:30]
		{
			tmp1 = (rxcas[chanid] >> 8) & 0x0f;
			if (tmp1 != pd->spans[spanid]->span.chans[15+chanid]->rxsig)
			{
				if(!(pd->spans[spanid]->span.chans[15+chanid]->sig & DAHDI_SIG_CLEAR))
				{
					dahdi_rbsbits(pd->spans[spanid]->span.chans[15+chanid], tmp1);
				}
			}
		}
	} 
}

static void rbs_delivery_quasar8(struct quasar * pq)
{
	struct quasar_dahdi *pd = pq->priv_data;
	unsigned int spanid, chanid, tmp1;

	/* Check signalisation */
	for (spanid=0; spanid < pd->num_spans; spanid++) {
		if (!(pd->spans[spanid]->span.flags & DAHDI_FLAG_RUNNING))
			continue;
		if (pd->spans[spanid]->span.lineconfig & DAHDI_CONFIG_CCS)
			continue;

			/* read interrupt status */
		if (framer_read(pq, spanid, 0x4, 0x1e) & 0x01) {
			for(chanid=1; chanid < 16; chanid++) {
				tmp1 = framer_read(pq, spanid, 0x9, 0x10 + chanid) & 0x0f;
				if (tmp1 != pd->spans[spanid]->span.chans[chanid-1]->rxsig) {
					if(!(pd->spans[spanid]->span.chans[chanid-1]->sig & DAHDI_SIG_CLEAR)) {
						dahdi_rbsbits(pd->spans[spanid]->span.chans[chanid-1], tmp1);
					}
				}
			}

			for(chanid=1; chanid < 16; chanid++) {
				/* chans > 16                            dahdi channels [17:31] -chans [16:30] */
				tmp1 = framer_read(pq, spanid, 0xa, 0x10 + chanid) & 0x0f;
				if (tmp1 != pd->spans[spanid]->span.chans[15+chanid]->rxsig) {
					if(!(pd->spans[spanid]->span.chans[15+chanid]->sig & DAHDI_SIG_CLEAR)) {
						dahdi_rbsbits(pd->spans[spanid]->span.chans[15+chanid], tmp1);
					}
				}
			}
		}
	}
}

/*
RX and TX buffers - 4096 bytes wide.
While DMA controller reads or writes
data from/to buffer addres is 
increasing from 0 to 4095.
 

 0  --------------------------
    I		0	     I
512 --------------------------
    I		1	     I
1024--------------------------
    I		2	     I
1536--------------------------
    I	        3            I
2048--------------------------
    I	    	4	     I
2560--------------------------
    I		5	     I
3072--------------------------
    I		6	     I
3584--------------------------
    I		7	     I
4095--------------------------
*/

static void copy_chunk_quasarm(struct quasar * pq) 
{
	struct quasar_dahdi *pd = pq->priv_data;
	unsigned int keep, block, read_offset, write_offset;
	unsigned int channels, offset, spanid, chanid;
	int i;

	keep = (pq->bar0[BAR0_DMACR] & 0xff000000) >> 24;
	if (keep != pd->ovfl_counter) {
		module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", pd->ovfl_counter);
		pd->ovfl_counter = keep;
	}
	
	keep = pq->bar0[BAR0_DMA_WADDR]; /* get relative addres of current dma operating position */

	/* Each block: 32 channels, 8 bytes (chunk), 2 channels */
	block = keep/32/DAHDI_CHUNKSIZE/2; /* Current block number */
	read_offset = (block+0x7)&0x7;	/* Read (Previous) block number */
	read_offset *= 32*DAHDI_CHUNKSIZE*2;
	write_offset = (block+0x1)&0x7;	/* Write (Next) block number */
	write_offset *= 32*DAHDI_CHUNKSIZE*2;

	//TODO: optimize a bit.
	i = 8 + (keep/512)-(pd->keep/512)-1;
	if (i >= 8) {
            pd->irqmisses += (int)i-8;
	} else {
            pd->irqmisses += (int)i;
	}
	pd->keep = keep;

	for (spanid=0; spanid < pd->num_spans; spanid ++) {
		if (pd->spans[spanid]->unframed) {
			channels = 32;
			offset = 0;
		} else {
			channels = 31;
			offset = 8;
		}
		for (chanid=0; chanid<channels; chanid++) { 
			pd->spans[spanid]->chans[chanid]->readchunk =  
				(unsigned char *) 
				((unsigned char *) pd->readchunk + 
				read_offset + 					/* block offset */
				offset +  						/* skip zero ts */
				chanid*DAHDI_CHUNKSIZE + 
				spanid * 32 * DAHDI_CHUNKSIZE); 

			pd->spans[spanid]->chans[chanid]->writechunk =  
				(unsigned char *) 
				((unsigned char *) pd->writechunk + 
				write_offset + 					/* block offset */
				offset +  						/* skip zero ts */
				chanid*DAHDI_CHUNKSIZE + 
				spanid * 32 * DAHDI_CHUNKSIZE); 
		}
	}

	if (0 == (pd->intcount & 127)) {
	    for (spanid=0; spanid < pd->num_spans; spanid++) {
			unsigned char ucTmp=0, ucTmp2=0;

			if (!pq->have_liu[spanid])
				continue;

			acc_lxt_reg(pq, spanid, LR_PSR, LIU_READ, &ucTmp);
			acc_lxt_reg(pq, spanid, LR_CR2, LIU_READ, &ucTmp2);
			if ((pd->spans[spanid]->psr ^ ucTmp) & LR_PSR_NLOOP) { /* NLOOP state changed */
			if (ucTmp & LR_PSR_NLOOP)
				module_printk(KERN_WARNING, "Network LoopUp detected @ Quasarm_%d_%d.\n", pq->dev_number, spanid);
		    else
		    	module_printk(KERN_WARNING, "Network LoopDown detected @ Quasarm_%d_%d.\n", pq->dev_number, spanid);
			}
			pd->spans[spanid]->psr = ucTmp;
	    }
	}

}

static void copy_chunk_quasar8(struct quasar * pq) 
{
	struct quasar_dahdi *pd = pq->priv_data;
	unsigned int keep, chunk, spanid, chanid, tmp1, tmp2, shift;


	/* check DMA overflow counter */
	keep = (pq->bar0[BAR0_DMACR] & 0xff000000) >> 24;
	if (keep != pd->ovfl_counter) {
		module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", pd->ovfl_counter);
		pd->ovfl_counter = keep;
	}
	
	keep = pq->bar0[BAR0_DMA_WADDR]; /* get relative addres of current dma operating position */

	if (keep < (DAHDI_CHUNKSIZE * 256)) {
		/* we are on the first part of buffer, serve second part of the buffer */
		if (pd->check == 1) {
			++pd->irqmisses;
			module_dprintk(KERN_DEBUG, "Miss chunk in 1st part.\n");
		}

		pd->check = 1;
		//	    if (keep > 0x80) printk("e12zap: Possible data loss=0x%x\n", keep);
		for (chunk = 0; chunk < DAHDI_CHUNKSIZE; chunk++) {
			for (spanid=0; spanid < pd->qn_span; spanid ++) { /* span */
				for (chanid=0; chanid<16; chanid++) { /* chan */
					shift = chanid + 16 * spanid +  64 * chunk + DAHDI_CHUNKSIZE * 64;
					tmp2 = pd->readchunk[shift];
					tmp1 = 0;
					if (chanid==0) { /* for 1 timeslot */
						/* transmit data from dahdi 1ts */
						if (pd->num_spans > 4) {
							pd->spans[4+spanid]->span.chans[0]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (pd->spans[4+spanid]->span.chans[0]->writechunk[chunk] << 16);
						}
						pd->spans[spanid]->span.chans[0]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |=  pd->spans[spanid]->span.chans[0]->writechunk[chunk];

					} else {
						/* transmit data from dahdi 2-31 ts */
						if (pd->num_spans > 4) {
							pd->spans[4+spanid]->span.chans[chanid*2-1]->readchunk[chunk] = tmp2 >> 24;
							pd->spans[4+spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (pd->spans[4+spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 24);
							tmp1 |= (pd->spans[4+spanid]->span.chans[chanid*2]->writechunk[chunk] << 16);

						}
						pd->spans[spanid]->span.chans[chanid*2-1]->readchunk[chunk] = 0xff & (tmp2 >> 8);
						pd->spans[spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |= (pd->spans[spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 8);
						tmp1 |= (pd->spans[spanid]->span.chans[chanid*2]->writechunk[chunk]);
					}
					pd->writechunk[shift] = tmp1;
				}
			}
		}

	} else {
		/* we are in second part of the buffer, serve first part of the buffer */

		if (pd->check == 2) {
			++pd->irqmisses;
			module_dprintk(KERN_DEBUG, "Miss chunk in 2nd part.\n");
		}
		pd->check = 2;
		//	    if (keep > 0x880) printk("e12zap: Possible data loss=0x%x\n", keep);
		for (chunk = 0; chunk < DAHDI_CHUNKSIZE; chunk++) {
			for(spanid=0; spanid < pd->qn_span; spanid ++) { /* span */
				for (chanid=0; chanid < 16 ;chanid++) { /* chan */
					shift = chanid + 16 * spanid +  64 * chunk;
					tmp2 = pd->readchunk[shift];
					tmp1 = 0;
					if (chanid==0) { /* for 1 timeslot */
						/* transmit data from dahdi 1ts */
						if (pd->num_spans > 4) {
							pd->spans[4+spanid]->span.chans[0]->readchunk[chunk] = 0xff & (tmp2 >> 16);
							tmp1 =  (pd->spans[4+spanid]->span.chans[0]->writechunk[chunk] << 16);
						}
						pd->spans[spanid]->span.chans[0]->readchunk[chunk] = 0xff & tmp2;
						tmp1 |=  pd->spans[spanid]->span.chans[0]->writechunk[chunk];

					} else {
						/* transmit data from dahdi 2-31 ts */
						if (pd->num_spans > 4) {
							pd->spans[4+spanid]->span.chans[chanid*2-1]->readchunk[chunk] = tmp2 >> 24;
							pd->spans[4+spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & (tmp2 >> 16);

							tmp1 =  (pd->spans[4+spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 24);
							tmp1 |= (pd->spans[4+spanid]->span.chans[chanid*2]->writechunk[chunk] << 16);
						}
						pd->spans[spanid]->span.chans[chanid*2-1]->readchunk[chunk] = 0xff & (tmp2 >> 8);
						pd->spans[spanid]->span.chans[chanid*2]->readchunk[chunk] = 0xff & tmp2;

						tmp1 |= (pd->spans[spanid]->span.chans[chanid*2-1]->writechunk[chunk] << 8);
						tmp1 |= (pd->spans[spanid]->span.chans[chanid*2]->writechunk[chunk]);
					}
					pd->writechunk[shift] = tmp1;
				}
			}
		}
	}

}

static void copy_chunk_quasar16(struct quasar * pq) 
{
	struct quasar_dahdi *pd = pq->priv_data;
	unsigned int keep, spanid, chanid;
	unsigned int buffer_offset, channels, offset;


	/* check DMA overflow counter */
	keep = (pq->bar0[BAR0_DMACR] & 0xff000000) >> 24;
	if (keep != pd->ovfl_counter) {
		module_dprintk(KERN_DEBUG, "DMA OVFL:%d\n", pd->ovfl_counter);
		pd->ovfl_counter = keep;
	}
	
	keep = pq->bar0[BAR0_DMA_WADDR]; /* get relative addres of current dma operating position */

	if (keep < (DAHDI_CHUNKSIZE * 16 * 32)) buffer_offset = DAHDI_CHUNKSIZE * 16 * 32;
		else buffer_offset = 0;

	for (spanid=0; spanid < pd->num_spans; spanid ++) {
		if (pd->spans[spanid]->unframed) {
			channels = 32;
			offset = 0;
		} else {
			channels = 31;
			offset = 8;
		}
		for (chanid=0; chanid<channels; chanid++) { 
			pd->spans[spanid]->chans[chanid]->readchunk =  
				(unsigned char *) ((unsigned char *) pd->readchunk + 
					buffer_offset + 										// half of buffer
					offset +  												// skip zero ts
					chanid*DAHDI_CHUNKSIZE + 
					spanid * 32 * DAHDI_CHUNKSIZE); 
		}
	}
}
int dacs_quasar16 (struct dahdi_chan *dst, struct dahdi_chan *src) {
	module_dprintk(KERN_DEBUG, "dahdi_dacs: not supported in quasar16\n");
	return -1;
}
int dacs_quasar2012 (struct dahdi_chan *dst, struct dahdi_chan *src) {
	struct quasar_span *qspan_dst = dst->pvt;
	struct quasar_dahdi *pd = qspan_dst->pvt;
	struct quasar *pq = pd->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");

	if (!dst)
		return -1;

	if (src) {
		struct quasar_span *qspan_src = src->pvt;
		if(qspan_dst->pvt != qspan_src->pvt) {
			/* trying to cross-connect timeslots from different boards */
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");

			if (qspan_dst->spannumber < 4) {
				pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(1, (dst->chanpos + 32 * qspan_dst->spannumber));
			} else {
				pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(3, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
			}
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", qspan_dst->spannumber, dst->chanpos, qspan_src->spannumber, src->chanpos);
		/* cross-connecting
		 4 -> 6
		 5 -> 7 */
		if (qspan_src->spannumber < 4) {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(0, (src->chanpos + 32 * qspan_src->spannumber));
		} else {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(2, (src->chanpos + 32 * (qspan_src->spannumber - 4)));
		}
	} else {
		/* return to default crossing */
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", qspan_dst->spannumber, dst->chanpos);

		if (qspan_dst->spannumber < 4) {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(1, (dst->chanpos + 32 * qspan_dst->spannumber));
		} else {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(3, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
		}
	}
	return 0;
}
int dacs_quasarm(struct dahdi_chan *dst, struct dahdi_chan *src)
{
	struct quasar_span * dst_span = dst->pvt;
	struct quasar_dahdi * pd = dst_span->pvt;
	struct quasar * pq = pd->pvt;

	if (!dst) return -1;
	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");
	if(src)
	{
		struct quasar_span * src_span = src->pvt;
		if(dst_span->pvt != src_span->pvt) //trying to cross-connect timeslots from different boards
		{
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");
			//here to return to default crossing
			//TO E1 ts	 //FROM DMA ts, 2 - is DMA buffer offset
			pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * dst_span->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE((2 + dst_span->spannumber), dst->chanpos);
			pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * src_span->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE((2 + src_span->spannumber), src->chanpos);
	
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", dst_span->spannumber, dst->chanpos, src_span->spannumber, src->chanpos);
		//TO E1 ts         											 //FROM E1 ts
		pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * dst_span->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE(src_span->spannumber, src->chanpos);
		pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * src_span->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE(dst_span->spannumber, dst->chanpos);
		return 0;			
	}
	else
	{
	    //return to default crossing
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", dst_span->spannumber, dst->chanpos);
		//TO E1 ts												 //FROM DMA ts, 2 - is DMA buffer offset
		pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * dst_span->spannumber) + dst->chanpos] = INPUT_TS_SIGNATURE((2 + dst_span->spannumber), dst->chanpos);
		//pq->bar0[SWITCH_TABLE_OFFSET + 0x1000 + (0x100 * src_span->spannumber) + src->chanpos] = INPUT_TS_SIGNATURE(2 + src_span->spannumber, src->chanpos);
		
		return 0;
	}
}

int dacs_quasar8 (struct dahdi_chan *dst, struct dahdi_chan *src) {
	struct quasar_span *qspan_dst = dst->pvt;
	struct quasar_dahdi *pd = qspan_dst->pvt;
	struct quasar *pq = pd->pvt;

	module_dprintk(KERN_DEBUG, "e12zap_dacs:\n");

	if (!dst)
		return -1;

	if (src) {
		struct quasar_span *qspan_src = src->pvt;
		if(qspan_dst->pvt != qspan_src->pvt) {
			/* trying to cross-connect timeslots from different boards */
			module_dprintk(KERN_DEBUG, "Can't switch timeslots from different boards, using soft switch\n");

			if (qspan_dst->spannumber < 4) {
				pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * qspan_dst->spannumber));
			} else {
				pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
			}
			return -1;
		}
		module_dprintk(KERN_DEBUG, "Switching %d:%d(in)/%d:%d(out).\n", qspan_dst->spannumber, dst->chanpos, qspan_src->spannumber, src->chanpos);
		/* cross-connecting
		 4 -> 6
		 5 -> 7 */
		if (qspan_src->spannumber < 4) {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(6, (src->chanpos + 32 * qspan_src->spannumber));
		} else {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(7, (src->chanpos + 32 * (qspan_src->spannumber - 4)));
		}
	} else {
		/* return to default crossing */
		module_dprintk(KERN_DEBUG, "Return to default crossing on %d:%d\n", qspan_dst->spannumber, dst->chanpos);

		if (qspan_dst->spannumber < 4) {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(4, (dst->chanpos + 32 * qspan_dst->spannumber));
		} else {
			pq->bar0[0x1000 + 0x100 * qspan_dst->spannumber + dst->chanpos] = CTB_TO_E1_TABLE(5, (dst->chanpos + 32 * (qspan_dst->spannumber - 4)));
		}
	}
	return 0;
}

struct quasar_hal_ops quasar8_hal_ops = {
	.init_core_regs = init_core_regs_quasar8,
	.init_switch_map = init_switch_map_quasar8,
	.set_hw_options = set_hw_options,
	.init_framer_regs = init_framer_regs_quasar8,
	.ioctl = ioctl_quasar8,
	.quasar_find_master = quasar_find_master,
	.check_alarm = check_alarm_quasar8,
	.clearstat = clearstat_quasar8,
	.maint = maint_quasar8,
	.rbsbits = rbsbits_quasar8,
	.shutdown = shutdown_quasar,
	.startup = startup_quasar8,
	.stop_dma = stop_dma,
	.rbs_delivery = rbs_delivery_quasar8,
	.copy_chunk = copy_chunk_quasar8,
	.dacs = 	dacs_quasar8,
};
struct quasar_hal_ops quasarm_hal_ops = {
	.init_core_regs = init_core_regs_quasarm,
	.init_switch_map = init_switch_map_quasarm,
	.set_hw_options = set_hw_options,
	.init_framer_regs = init_framer_regs_quasarm,
	.ioctl = ioctl_quasarm,
	.quasar_find_master = quasar_find_master,
	.check_alarm = check_alarm_quasarm,
	.clearstat = clearstat_quasarm,
	.maint = maint_quasarm,
	.rbsbits = rbsbits_quasarm,
	.shutdown = shutdown_quasar,
	.startup = startup_quasarm,
	.stop_dma = stop_dma,
	.rbs_delivery = rbs_delivery_quasarm,
	.copy_chunk = copy_chunk_quasarm,
	.dacs = 	dacs_quasarm,
};

struct quasar_hal_ops quasar2012_hal_ops = {
	.init_core_regs = init_core_regs_quasar2012,
	.init_switch_map = init_switch_map_quasar2012,
	.set_hw_options = set_hw_options,
	.init_framer_regs = init_framer_regs_quasarm,
	.ioctl = ioctl_quasarm,
	.quasar_find_master = quasar_find_master,
	.check_alarm = check_alarm_quasarm,
	.clearstat = clearstat_quasarm,
	.maint = maint_quasarm,
	.rbsbits = rbsbits_quasarm,
	.shutdown = shutdown_quasar,
	.startup = startup_quasarm,
	.stop_dma = stop_dma,
	.rbs_delivery = rbs_delivery_quasarm,
	.copy_chunk = copy_chunk_quasar8,
	.dacs = 	dacs_quasar2012,
};
struct quasar_hal_ops quasar16_hal_ops = {
	.init_core_regs = init_core_regs_quasar16,
	.init_switch_map = init_switch_map_quasar16,
	.set_hw_options = set_hw_options,
	.init_framer_regs = init_framer_regs_quasar16,
	.ioctl = ioctl_quasarm,
	.quasar_find_master = quasar_find_master,
	.check_alarm = check_alarm_quasarm,
	.clearstat = clearstat_quasarm,
	.maint = maint_quasarm,
	.rbsbits = rbsbits_quasarm,
	.shutdown = shutdown_quasar,
	.startup = startup_quasarm,
	.stop_dma = stop_dma,
	.rbs_delivery = rbs_delivery_quasarm,
	.copy_chunk = copy_chunk_quasar16,
	.dacs = 	dacs_quasar16,
};
